/************************************************************************************************************************/
/*

Software for smart cameras from Vision Components
-------------------------------------------------

Program: Klaus Schneider, VC


Program Description:
--------------------

*/
/************************************************************************************************************************/

#define NEW_IMAGE_VAR

#include <register.h>
#include <vcrt.h>
#include <flib.h>
#include <vclib.h>
#include <macros.h>
#include <sysvar.h>

#define _TI_ENHANCED_MATH_H
#include <math.h>
#include <mathf.h>
#include <string.h>
#include <stdlib.h>     /* atof print("%f\n", (float)atof("-1.345")); */

/************************************************************************************************************************/

// VCLibs licence code
#include "\ti\vclib\licence.h"

/************************************************************************************************************************/

// define file name for calibration and parameter file on flash

#define  MAX_SETTING_SIZE        40000
#define  MAX_PROD_RESULT_STRING   2000

#define  SENSOR_FILE        "fd:/VC3DPar.001"    // sensor parameter file name
#define  PRODUCT_FILE       "fd:/ProdPar.001"    // product parameter file name

/************************************************************************************************************************/

////////////////////////////////////////////////////////////////////////////////////////
// TESTING VERSION with telnet output, auto trigger and module display 
////////////////////////////////////////////////////////////////////////////////////////

// #define   TESTING_VERSION
#define   CHECK_LICENCE_CODE

/************************************************************************************************************************/

// adjust s timer 0
#define MAX_JITTER_US         50  // max. allowed auto trigger jitter [s] until increasing AutoTriggerError
#define TIMER0_US_DIV         25  // on 62xx nano cameras 1 ms / 25000 = 40ns per dig  ->  25 x 40ns = 1s
#define TIMER0_DIG            (*(volatile U32 *)(0x01C21400 + 0x10))

/************************************************************************************************************************/

// sensor defines

// max. header size of TCP/IP communication
#define GLOBAL_HEAD_SIZE      32
#define GLOBAL_HEAD_SYNC      0xAA0F55F0   /* = 10101010000011110101010111110000 */

// Lib sub pixel accuracy (SubPixel = Position / LASER_POS_MUL)
#define LASER_POS_MUL         64

// z coordinate, if no laser is detected
#define NO_LASER_DETEC      (-100 * LASER_POS_MUL)

/************************************************************************************************************************/

// product defines
#define MAX_JOBS               8
#define MAX_POLYGON_POINTS     16

/************************************************************************************************************************/

// Host command defines

#define CMD_ERROR	                       -1
#define CMD_STOP	                        0
#define CMD_LASER_MODE	                  1
#define CMD_ROI_X	                        2
#define CMD_ROI_Y	                        3
#define CMD_ROI_DX	                      4
#define CMD_ROI_DY	                      5
#define CMD_GAIN	                        6
#define CMD_SHUTTER	                      7
#define CMD_TRIGGER	                      8
#define CMD_MODE 	                        9
#define CMD_NBR_LINES	                   10
#define CMD_METHOD	                     11
#define CMD_ALGORITHM	                   12  // not used any more 
#define CMD_THRESH                       13
#define CMD_MINWIDTH                     14
#define CMD_MAXWIDTH                     15
#define CMD_EXPMODE                      16
#define CMD_LASERVAL                     17
#define CMD_MINSH                        18
#define CMD_MAXSH                        19
#define CMD_DOUBLE_SH1                   20
#define CMD_DOUBLE_SH2                   21
#define CMD_FILTER                       22
#define CMD_ROI_X_MM                     23
#define CMD_ROI_Z_MM                     24
#define CMD_ROI_DX_MM                    25
#define CMD_ROI_DZ_MM                    26
#define CMD_LIB_VERSION                  27
#define CMD_PLC_IN                       28
#define CMD_PLC_OUT                      29
#define CMD_ETHERNET_PACKAGE             30
#define CMD_AUTO_TRIG_PERIODE            31
#define CMD_AUTO_TRIG_ERR                32
#define CMD_SEND_NOWAIT                  34
#define CMD_SUBSAMPLE                    35
#define CMD_WORLD_ACCURACY               36  // not used any more
#define CMD_SEND_INFO_PAR                37  // not used any more
#define CMD_CAMERA_SNR                   39

#define CMD_SAVE_PRODUCT_FILE            45  // save product parameter on camera file "fd:/ProdPar.001"
#define CMD_LOAD_PRODUCT_FILE            38  // load product parameter from camera file "fd:/ProdPar.001"
#define CMD_RECEIVE_PRODUCT_DATA         40  // product parameter via TCP/IP: Host ->  Cam
#define CMD_SEND_PRODUCT_DATA            41  // product parameter via TCP/IP: Cam  -> Host

#define CMD_SAVE_SENSOR_FILE             33  // save sensor parameter on camera file "fd:/VC3DPar.001"
#define CMD_LOAD_SENSOR_FILE             42  // load sensor parameter from camera file "fd:/VC3DPar.001"
#define CMD_RECEIVE_SENSOR_DATA          43  // sensor parameter via TCP/IP: Host ->  Cam
#define CMD_SEND_SENSOR_DATA             44  // sensor parameter via TCP/IP: Cam  -> Host

#define CMD_SPECKLE_FILTER               46
#define CMD_MEDIAN_FILTER                47

/************************************************************************************************************************/

#define  EVEN(x)               ((x)&0xFFFFFFFE)
#define  Round(x)              ((I32)floorf(x+0.5))
#define  RoundI40(x)           ((I40)floorf(x+0.5))

/************************************************************************************************************************/

////////////////////////////////////////////////////
// error messages
////////////////////////////////////////////////////

// #define ERR_NONE               0

#define ERR_MEM_INI        -1000
#define ERR_LICENSE_CODE   -2000

#define ERR_FILE_OPEN      -2100
#define ERR_FILE_SIZE      -2200
#define ERR_FILE_DATA      -2300
#define ERR_FILE_VERS      -2400

#define ERR_TCP_IP         -3001
#define ERR_COUNTER        -3002
#define ERR_PARM           -3003
#define ERR_COMMAND        -3004
#define ERR_MEM_RLC        -3005
// #define ERR_NO_SET_CMD     -3006
// #define ERR_NO_READ_CMD    -3007
#define ERR_TCP_IP_SYNC    -3008
#define ERR_PARM_NAME      -3009
#define ERR_CMD_STRG_LEN   -3010
#define ERR_CMD_TYPE       -3011
#define ERR_CMD_DATA_LEN   -3012
#define ERR_CMD_LESS_DATA  -3013
#define ERR_CMD_SOCKET     -3014
#define ERR_CMD_RECEIVE    -3015
#define ERR_PARM_RANGE_MIN -3016
#define ERR_PARM_RANGE_MAX -3017
#define ERR_MEM_PROD_ANGLE -3018

#define ERR_PROD_INACTIV   -4000  // job was not executed
#define ERR_PROD_ROI_NR    -4001  // less than expected ROI points
#define ERR_PROD_ANGLE     -4002  // no line in selected angle range
#define ERR_PROD_LINE_NR   -4003  // less than expected points for line
#define ERR_PROD_GENERAL   -4004  // general product calculation error
#define ERR_PROD_RESULT    -4005  // couldn't calculate product result


////////////////////////////////////////////////////
// define modes
////////////////////////////////////////////////////

// parameter setting mode 
enum ParMode {PARAMETER_LOAD, PARAMETER_SAVE, PARAMETER_SEND, PARAMETER_RECEIVE, PARAMETER_SHOW};

// screen buffer mode 
enum ScrMode {SCREEN_BUFF_INI, SCREEN_BUFF_SWAP, SCREEN_BUFF_RESET, SCREEN_BUFF_DEINI, SCREEN_BUFF_FIRST};

// send buffer mode 
enum SendMode {INI_SEND_BUFFER, USE_SEND_BUFFER, EMPTY_SEND_BUFFER};

// auto trigger mode and result 
enum TrigMode {AUTO_TRIG_INI, AUTO_TRIG_DIFF, AUTO_TRIG_READY};

// parameter ini
#define MEM_INI                0
#define MEM_DEINI              1

// screen buffer result
#define SCREEN_RES_END         0
#define SCREEN_RES_IMAGE       1
#define SCREEN_RES_INI         2
#define SCREEN_RES_BUSY       -1
#define SCREEN_RES_ERR_PIPE   -2

// exposure mode
#define FIX_EXPOSURE           0
#define AUTO_EXPOSURE          1
#define DOUBLE_EXPOSURE        2

// TRIGGER mode
#define AUTO_TRIGGER_NEG      -3
#define AUTO_TRIGGER_POS      -2
#define AUTO_TRIGGER_MODE     -1
#define NONE_TRIGGER_MODE      0
#define TRIGGER_RISING_EDGE    1
#define TRIGGER_FALLING_EDGE   2

/************************************************************************************************************************/

typedef struct
{
  ////////////////////////////////////////////////////////////////
  // version
  ////////////////////////////////////////////////////////////////

  I32   MainVersion;
  I32   LibVersion;
  I32   CameraSNR;
  
  ////////////////////////////////////////////////////////////////
  // licence parameter
  ////////////////////////////////////////////////////////////////

  // please set LicenceCode to the value offered from Vision Components
  U32   LicenceCode1;    // hex value like 0x1234ABCD
  U32   LicenceCode2;    // hex value like 0xABCD1234
  
  
  ////////////////////////////////////////////////////////////////
  // ASCII sensor setting parameter file
  ////////////////////////////////////////////////////////////////

  I32   SensorSettingSize;
  I8    *pSensorSetting;
  

  ////////////////////////////////////////////////////////////////
  // sensor parameter
  ////////////////////////////////////////////////////////////////
  
  
  // reset 
  I32   Stop;
  
  // laser on / off / pulse
  I32   Laser;
  
  // define ROI
  I32   ROI_X;
  I32   ROI_Y;
  I32   ROI_DX;
  I32   ROI_DY;  
  I32   ROI_X_MM;
  I32   ROI_Z_MM;
  I32   ROI_DX_MM;
  I32   ROI_DZ_MM;
  
  // sensor gain (high gain lead to additional noise)
  I32   GainVal;
  
  // shutter time in s
  I32   Shutter;
  I32   ShutterReal;
  
  // trigger mode 
  I32   Trigger;
  
  // result transfer mode
  I32   DataMode;
  
  // number scanning lines
  I32   NbrLines;
  
  // laser detection mode
  I32   Methode;
  
  // float or I16 transfer type
  I32   Algorithm;

  // rlc fast laser detection parameter
  I32   RlcThresh;

  // sub sample for model 6211 (yr models)
  I32   SubSample;
  I32   ROI_Y_SubSample;
  I32   ROI_DY_SubSample;
  
  // laser sub pixel detection parameter
  I32   StepDx;
  I32   MinWidth;
  I32   MaxWidth;
  
  // Shutter Mode (FIX_EXPOSURE, AUTO_EXPOSURE, DOUBLE_EXPOSURE)
  I32   ExposureMode;
  
  // autoshutter
  I32   AutoShutterVal;
  I32   AutoShutterFilter;
  I32   AutoShutterMin;
  I32   AutoShutterMax;

  // double exposure
  I32   DoubleShutterImg;
  I32   DoubleShutter[2];
  
  // filter
  I32   ReflexionFilter;
  I32   MedianFilter;
  I32   SpeckleFilter;
  I32   SkipNoLaserPoints;
  I32   LaserMaskFilter;

  // 0: wait until TCP/IP transfer is complete / 1: send as much as possible
  I32   EthernetSendNoWait;

  // pack nr lines together for more Ethernet speed (0 = use max. memory) 
  I32   EthernetPackNr;
  
  // auto trigger mode (active if TRIGGER = AUTO_TRIGGER_MODE)
  I32   AutoTriggerFPS;
  I32   AutoTriggerError; 
  
  // command transfer mode (0=Binary / 1=ASCII)
  I32   CommandModeASCII;
  
  // PCL settings
  I32   InPLC;
  I32   OutPLC;
  
  // time stamp for image exposer ready time in ms (since camera boot)
  U64   ImgTimeStampMs;

  
  ////////////////////////////////////////////////////////////////
  // system calibration values
  ////////////////////////////////////////////////////////////////

  // Tsai Parameter
  I32   CalibrationErrorTsai;
  
  // Kanatani Parameter
  I32   CalibrationErrorKanatani;
  I8   *KanataniErrorMessage;
  

  ////////////////////////////////////////////////////////////////
  // general parameters
  ////////////////////////////////////////////////////////////////

  // debug result
  I32   PrintHostCommand;
  I32   DebugPrint;
  I32   DebugDraw;
  
  // TCP/IP Receive Data Buffer
  I32   MinReceiveBuff;
  I32   MaxReceiveBuff;
  U8    *ReceiveBuffer[2];

  // TCP/IP Send Data Buffer
  I32   MaxSendBuff;
  U8    *SendBuffer[2];


  ////////////////////////////////////////////////////////////////
  // laser detection results (absolute positions)
  ////////////////////////////////////////////////////////////////
  
  // all nr of laser detection positions (columns)
  I32   NrLaserPositions;

  // top (Z0) and bottom (Z1) position of laser (last image detection, pixel accuracy)
  I32   *pLaserPosZ0;
  I32   *pLaserPosZ1;
  
  // laser position at double shutter
  I32   *pLaserPosZ_Sh0;
  I32   *pLaserPosZ_Sh1;
  U8    *pLaserVal_Sh0;
  U8    *pLaserVal_Sh1;


  ////////////////////////////////////////////////////////////////
  // final laser detection results (absolute positions)
  ////////////////////////////////////////////////////////////////

  // valid nr of laser detection points
  I32   LaserPoints;
  
  // laser brightness
  U8    *pLaserVal;
  
  // laser brightness at ROIs
  I32   NrLaserVal_ROIs;
  U8    *pLaserVal_ROIs;
  
  // accurate laser position IMAGE coordinates (sub pixel accuracy factor LASER_POS_MUL)
  I32   *pLaserPosX; 
  I32   *pLaserPosZ;

  // accurate laser position WORLD coordinates
  float *pLaserPosX_MM; 
  float *pLaserPosZ_MM;
  
  // result
  I32   Error;
} 
LsParameter;

/************************************************************************************************************************/

typedef struct
{
  /////////////////////////////////////////////////////////////////////////////////
  // General parameter
  /////////////////////////////////////////////////////////////////////////////////

  // product version number
  I32   ProductVersion;
  
  // pointer to product setting information from file (MAX_PARAMETER_SIZE)
  I32   ProductFileError;
  I32   ProdSettingSize;
  I8    *pProductSetting;
  
  // adjust Sensor size to ROIs (0=manual 1=auto 2=full)
  I32   AdjustSensorRoi;  


  /////////////////////////////////////////////////////////////////////////////////
  // Input Parameter
  /////////////////////////////////////////////////////////////////////////////////

  // ProductType 0: Product_AngleMeasure
  I32   ProductType[MAX_JOBS];
  
  // number polygon points (two points for rectangle area, more points for polygon, <2 points lead to error message)
  I32   PolygonNr[MAX_JOBS];

  // world coordinates
  point PolygonPoint[MAX_JOBS][MAX_POLYGON_POINTS];

  // Bounding box of the polygon areas for faster calculation
  float PolMinX[MAX_JOBS];
  float PolMinY[MAX_JOBS];
  float PolMaxX[MAX_JOBS];
  float PolMaxY[MAX_JOBS];
  

  ///////////////////////////////////////////////
  // ProductType 1: AngleMeasure
  ///////////////////////////////////////////////

  // 0=BestLine 1=HistoLine
  I32   LineAlgoMode[MAX_JOBS];
  
  // HistoLine Filter
  I32   HistoLineFilter[MAX_JOBS];

  // setting values
  float TargetAngle[MAX_JOBS];
  float AngleTolPos[MAX_JOBS];
  float AngleTolNeg[MAX_JOBS];
  
  // limit measure angel [0..180]
  float PhiMin[MAX_JOBS];
  float PhiMax[MAX_JOBS];

  // min. number detection points for valid lines
  I32   MinNrLaserPoints[MAX_JOBS];


  ///////////////////////////////////////////////
  // ProductType 1000: Product_Calibration
  ///////////////////////////////////////////////

  // define customer parameter
  I32   MatrixPosLeft;    // 0: from right / 1: from left
  float MaxPointDisMM;    // max. allowed distance for valid points to the matrix line 
  float MaxLineGabMM;     // max. allowed gab between valid points inside the matrix line 
  float ROI_Dis_MM;       // dx gab between matrix and bending ROI
  float ROI_Width_MM;     // bending ROI size dx
  float ROI_DIFF_ALPHA;   // bending ROI angle alpha to define bending ROI (lower left corner)
  float ROI_DIFF_BETA;    // bending ROI angle beta to define bending ROI (upper left corner)

  
  ///////////////////////////////////////////////
  // ProductType 1001: Advanced Line Detection
  ///////////////////////////////////////////////
  
  I32   LineDetectionMode [MAX_JOBS];
  I32   MaxHoughLineNr    [MAX_JOBS];
  I32   MinHoughPointDis  [MAX_JOBS];
  I32   MaxHoughPointDis  [MAX_JOBS];
  float RotationPointX_MM [MAX_JOBS];
  float RotationPointZ_MM [MAX_JOBS];
  float MaxLineDisSelect  [MAX_JOBS];
  float MaxLineDisBestFit [MAX_JOBS];
  
  
  /////////////////////////////////////////////////////////////////////////////////
  // Output Parameter
  /////////////////////////////////////////////////////////////////////////////////
  
  // output string
  I32   ProductResultLen;
  I8   *ProductResultString;
  
  I32   ProductError[MAX_JOBS];
  I32   Error;
  

  ///////////////////////////////////////////////
  // ProductType 1: AngleMeasure
  ///////////////////////////////////////////////

  // line parameters. Angle [0..360[, 0 degree at 3 o'clock, running clockwise
  float ResultAngle[MAX_JOBS];

  float LineCx[MAX_JOBS];
  float LineCy[MAX_JOBS];
  float LineB [MAX_JOBS];
}
ProductParameter;

/************************************************************************************************************************/

// Laser Scanner Lib Functions

I32  ScannerInit         (LsParameter *LsPar, I32 Mode);
I32  ScannerDetection    (LsParameter *LsPar, image *ImgArea);
I32  GetCameraID         (void);

void DelNoneDetectionPoints (LsParameter *LsPar);

void ImgToWorldCoord     (LsParameter *LsPar, I32 Nr, I32 *restrict pImgX, I32 *restrict pImgZ, float *restrict pWorldX, float *restrict pWorldZ);
void WorldToImgCoord     (LsParameter *LsPar, I32 Nr, float *restrict pWorldX, float *restrict pWorldZ, float *restrict pImgX, float *restrict pImgZ);  
void GetMaxSize          (LsParameter *LsPar, I32 ImgOrWorld, float *MinX, float *MaxX, float *MinY, float *MaxY);

void ImgToWorld_ROI      (LsParameter *LsPar);
void WorldToImg_ROI      (LsParameter *LsPar);

void LimitToMaxSize      (LsParameter *LsPar, I32 ImgOrWorld, float *MinX, float *MaxX, float *MinY, float *MaxY);

void SensorReset         (LsParameter *LsPar);

FILE *OpenFile           (char *filename, int debug, int write);
void CloseFile           (FILE *fp);
I32  ReadFileToMem       (char *FileName, char *pData, I32 MaxData, I32 debug);
I32  WriteMemToFile      (char *FileName, char *pData, I32 NrData, I32 debug);
I32  DeleteComment       (char *pDataIn, I32 NrIn, char *pDataOut);
I32  ReadIntValue        (char *pData, char *VarName, I32 *value, I32 debug);
I32  ReadFloatValue      (char *pData, char *VarName, float *value, I32 debug);
I32  ReadIntVar          (char *pData, char *VarName, I32 def, I32 min, I32 max, I32 *value, I32 debug);
I32  ReadFloatVar        (char *pData, char *VarName, float def, float min, float max, float *value, I32 debug);
I32  WriteIntToFile      (FILE* fp, char* setting, int value);
I32  WriteFloatToFile    (FILE* fp, char* setting, float value);
char *FindStringPosEnd   (char *restrict pData, char *needle);

I32  AdjustProductParameter(LsParameter *LsPar);
I32  GetAvgMax_U32P      (U32 *line, I32 nr, I32 filter, float *Pos, U32 *strength);

/************************************************************************************************************************/
