/************************************************************************************************************************/
/*

Software for smart cameras from Vision Components
-------------------------------------------------

Program: Klaus Schneider, VC

VC 3D Laser Scanner Demo

The program runs as a standalone program on a VC62xx camera.
For 3D result display or parameter settings use the PC Host Software VC3D.exe.

Main functions are:

I32  ScannerInit       (LsParameter *LsPar, I32 Mode);
I32  ScannerDetection  (LsParameter *LsPar, image *ImgArea);
void ImgToWorldCoord   (LsParameter *LsPar, I32 Nr, float *restrict pImgX, float *restrict pImgZ, float *restrict pWorldX, float *restrict pWorldZ);

Call once at the beginning of the program the function ScannerInit(LsPar, MEM_INI) and at the end of your program the function ScannerInit(LsPar, MEM_DEINI)
in order to release memory.

The program is designed for the VCLIB 3.06 and higher.

*/
/************************************************************************************************************************/

#include "LaserScanner_Header.h"

/************************************************************************************************************************/

#define MAIN_VERSION  559
#define MAIN_HOT_FIX    4

// #define PRODUCT_VERSION

/***************************************************************************************************************************/

// TCP / IP
#define TCP_TIMEOUT          500    // ms

/************************************************************************************************************************/

#ifdef PRODUCT_VERSION
extern ProductParameter ProdPar[1];
#endif

/************************************************************************************************************************/

///////////////////////////////////////////////////////////////////////////////////////////////////////
// TCP / IP Functions
///////////////////////////////////////////////////////////////////////////////////////////////////////

I32  TCP_NewConnection       (U32 Sock, U32 *Alias);
I32  TCP_ReadCommand         (LsParameter *LsPar, U32 *Alias, U8 *Byte, I32 minlen, I32 maxlen);
I32  TCP_RunCommand          (LsParameter *LsPar, U8 *pCommand, I32 CmdLen, U32 Alias);
I32  TCP_SendCommand         (LsParameter *LsPar, U32 Alias, U8 *Command, I32 maxlen, enum SendMode Mode);
I32  TCP_ReadData            (U32 *Alias, U8 *pData, I32 MinLen, I32 MaxLen, I32 TimeOutMs, I32 *DataLen);

I32  TCP_Open                (I32 Port, U32 *NewSock);
I32  TCP_Accept              (U32 Sock, U32 *FromSock);
I32  TCP_Send                (U32 Sock, char *SendData, I32 SendBytes);
I32  TCP_Receive             (U32 Sock, char *ReceiveData, I32 ReceiveBytes);
I32  TCP_Close               (U32 Sock);


///////////////////////////////////////////////////////////////////////////////////////////////////////
// Function declarations
///////////////////////////////////////////////////////////////////////////////////////////////////////

// PRODUCT
I32  Product_Measure         (LsParameter *LsPar);
I32  ProductParameterIni     (LsParameter *LsPar, I32 Mode);
I32  ProductParameterExchange(LsParameter *LsPar, ProductParameter *ProdPar, enum ParMode Mode);
I32  ProductParameterRange   (LsParameter *LsPar, ProductParameter *ProdPar, void *pPar, I32 *SetVal, I32 *MinVal, I32 *MaxVal);
I32  SetProductParI32        (ProductParameter *ProdPar, I32 *pPar, I32 *MinVal, I32 *MaxVal, I32 *NewVal, I32 Action);
I32  SetProductParF32        (ProductParameter *ProdPar, F32 *pPar, I32 *MinVal, I32 *MaxVal, I32 *NewVal, I32 Action);

// SENSOR
I32  SensorParameterIni      (LsParameter *LsPar, I32 Mode);
I32  SensorParameterRange    (LsParameter *LsPar, void *pPar, I32 *SetVal, I32 *MinVal, I32 *MaxVal);
I32  SensorParameterExchange (LsParameter *LsPar, enum ParMode Mode);
void ChangeParameter         (LsParameter *LsPar);
void SensorParameterChange   (LsParameter *LsPar, I32 *ParValue, I32 Step, I8 *text);
void ChangeValue_Telnet      (I32 *ParValue, I32 Min, I32 Max, I32 Step, I8 *text);
int  ReadKeyboard_Telnet     (int wait);
I32  ReadyForCalc            (LsParameter *LsPar);
I32  ScreenBuffer2           (LsParameter *LsPar, enum ScrMode Mode);
I32  AutoShutter             (LsParameter *LsPar);
I32  AutoTrigger             (LsParameter *LsPar, enum TrigMode Mode);
I32  SwitchLaser             (I32 LaserOn);
I32  SetImageFormat          (LsParameter *LsPar);
I32  StoreImage              (void);
I32  ReadImage               (void);
U40  GetTimeMs               (void);
I32  SetGlobalHeader         (U8 *pWrite, I32 DataLength, I32 Error, I32 Debug);
I32  GetGlobalHeader         (U8 *pRead, I32 ReadLen, I32 *Error, I32 *TimeOutMs, I32 *CheckSumData, I32 *DataLength, I32 *Offset, I32 Debug);

I32  set_image_format_ev76c560_subs (I32 subsampling_factor, I32 hpos, I32 vpos, I32 hsize, I32 vsize, I40 shutter, I32 mode, I32 * sen_hpos, I32 * sen_vpos, I32 * sen_hsize, I32 * sen_vsize, I40 * sen_shutter);

/************************************************************************************************************************/

// global header counter for Receive and Send TCP/IP commands
U32 TcpRecveCnt [2];
U32 TcpSendCnt  [2];

// sub header result counter 
U32 SendCounter [2];

I32 DataMode    [2];
I32 NbrLines    [2];
U32 PN = 0;          // Port Number

extern I32 DisplHoughLine;
extern I32 OffsLaserWidth;

/************************************************************************************************************************/

void main(void)
{
  LsParameter   LsPar[1];
  I32           SERVER_PORT[2];

  I32           response=-1, RunMode = 1, AutoTrigMode = 0, Error = 0;
  U32           sock[2], alias[2];
  U40           MsD, Ms0=0, Ms1;


  /////////////////////////////////////////////////////////////////////////////////
  // set higher task priority
  /////////////////////////////////////////////////////////////////////////////////

  I32 oldprio, highprio = 7, id = task_get_id(); // valid priority range: 1=highest 18=lowest priority

  task_set_priority(id, highprio, &oldprio); // set task priority
  print("Set own task '%s' to prio %d! (oldprio=%d)\n", task_get_name_from_id(id), highprio, oldprio);
  

  /////////////////////////////////////////////////////////////////////////////////
  // set licence code for VC libraries
  /////////////////////////////////////////////////////////////////////////////////

  /*******************************************************************************/
  /*                 PLEASE INITIALIZE YOUR LICENCE CODES                        */
  /*                    to activate library functions                            */
  /*                                                                             */
  /* VCLIB :         init_licence("T1122334455") or init_licence("LT1122334455") */
  /* Extension Lib : init_licence("E1122334455") or init_licence("LE1122334455") */
  /* Color Lib :     init_licence("C1122334455") or init_licence("LC1122334455") */
  /* etc.                                                                        */
  /*                                                                             */
  /*                                                                             */

  /* OR if available use "licence.h"                                             */
  INIT_LICENCE();

  /*                                                                             */
  /*******************************************************************************/

  
  /////////////////////////////////////////////////////////////////////////////////
  // Set Licence code for all VC3D nano Laser Scanner Cameras. 
  // For using the lib on other VC cameras, please ask VC for a new licence code
  /////////////////////////////////////////////////////////////////////////////////

  LsPar->LicenceCode1      = 0x674F27A1;
  LsPar->LicenceCode2      = 0x2E05F267;

  
  /////////////////////////////////////////////////////////////////////////////////
  // version information
  /////////////////////////////////////////////////////////////////////////////////

  if(1)
  {
    I32 Version, SubVersion;

    print("\nLinked with following library versions:\n");

    // GetLibVer_FASTLIB (&Version, &SubVersion);
    // print("FASTLIB : %d_%d %s\n", Version, SubVersion, GetLibVerString_FASTLIB());

    GetLibVer_VCLIB (&Version, &SubVersion);
    print("VCLIB   : %d_%d %s\n", Version, SubVersion, GetLibVerString_VCLIB());

    GetLibVer_VCRTLIB (&Version, &SubVersion);
    print("VCRTLIB : %d_%d %s\n", Version, SubVersion, GetLibVerString_VCRTLIB());

    print("\nCompiled with following header versions:\n");

    print("FLIB.H  : %u.%d\n", FASTLIB_VERSION, FASTLIB_SUBVERSION);
    print("VCLIB.H : %u.%d\n", VCLIB_VERSION, VCLIB_SUBVERSION);
    print("VCRT.H  : %u.%d\n", VCRTLIB_VERSION, VCRTLIB_SUBVERSION);

    print("\nVC/RT OS   running  with Version: %u.%d\n\n", getvar(VERSION), getvar(TESTVERSION));

    print("Compiled at %s %s Camera=%d\n\n", __DATE__, __TIME__, getvar(MODEL));
  }


  /////////////////////////////////////////////////////////////////////////////////
  // ini camera
  /////////////////////////////////////////////////////////////////////////////////

  // stop live image mode
  tpict();

  // set trigger output to exposer mode
  TRIGOUT_EXP();

  // set trigger input to positive edge
  TRIGINP_POS();

  // set logical page to standard image acquisition page
  ScrSetLogPage((I32)ScrGetCaptPage);  


  /////////////////////////////////////////////////////////////////////////////////
  // ini scanner memory and lens calibration 
  /////////////////////////////////////////////////////////////////////////////////

  Error = ScannerInit(LsPar, MEM_INI);

  if(LsPar->CalibrationErrorKanatani == 0) print("Kanatani Calibration Ready.\n");
  if(LsPar->CalibrationErrorTsai     == 0) print("Tsai     Calibration Ready.\n");

  if(Error)
  {
    if(LsPar->CalibrationErrorKanatani)
    {
      print("Kanatani Initialisation Error %d: %s\n", LsPar->CalibrationErrorKanatani, LsPar->KanataniErrorMessage);
    }

    if(LsPar->CalibrationErrorTsai)
    {
      print("Tsai Initialisation Error: ");
      if(LsPar->CalibrationErrorTsai == ERR_FILE_OPEN) {print("Missing system calibration file.\n");}
      if(LsPar->CalibrationErrorTsai == ERR_FILE_SIZE) {print("Wrong system calibration file size\n");}
      if(LsPar->CalibrationErrorTsai == ERR_FILE_DATA) {print("Corrupted system calibration file\n");}
    }

    if(Error == ERR_MEM_INI) {print("Error: Memory error in ScannerInit()\n"); goto QUIT;}
  }

  print("\n");


  /////////////////////////////////////////////////////////////////////////////////
  // ini SENSOR parameter (after ScannerInit())
  /////////////////////////////////////////////////////////////////////////////////

  Error = SensorParameterIni(LsPar, MEM_INI);  
  print("\n");


  /////////////////////////////////////////////////////////////////////////////////
  // ini PRODUCT parameter (after ScannerInit())
  /////////////////////////////////////////////////////////////////////////////////
  
  #ifdef PRODUCT_VERSION  
  Error = ProductParameterIni(LsPar, MEM_INI);
  print("\n");    
  #endif


  /////////////////////////////////////////////////////////////////////////////////
  // allocate image buffer for parallel mode
  /////////////////////////////////////////////////////////////////////////////////

  Error = ScreenBuffer2(LsPar, SCREEN_BUFF_INI);
  if(Error == ERR_MEMORY) {print("Error %d: ScreenBuffer2()\n", Error); goto QUIT;}


  /////////////////////////////////////////////////////////////////////////////////
  // ini TCP/IP connection
  /////////////////////////////////////////////////////////////////////////////////

  SERVER_PORT[0] =  1096;    // scanner communication port 0 (master)
  SERVER_PORT[1] =  1097;    // scanner communication port 1 (slave)

  for(PN=0; PN<2; PN++) 
  {
    sock        [PN] = 0;
    alias       [PN] = 0;
    TcpRecveCnt [PN] = 0;
    TcpSendCnt  [PN] = 0;
    SendCounter [PN] = 0;
    DataMode    [PN] = 0;
    NbrLines    [PN] = 0;
  }

  // set ini values at server port 0
  DataMode    [0] = LsPar->DataMode;
  NbrLines    [0] = LsPar->NbrLines;
  
  for(PN=0; PN<2; PN++) 
  {
    print("Open Port %d\n", SERVER_PORT[PN]);

    response = TCP_Open(SERVER_PORT[PN], &sock[PN]);
    
    if (response ==  0) print("New stream socket created at port %d at SocketAdr=0x%x\n", SERVER_PORT[PN], sock[PN]);
    if (response == -1) print("Create stream socket failed\n");
    if (response == -2) print("Stream bind failed\n");
    if (response == -3) print("Listen failed\n");

    if (response !=  0) goto QUIT;

    print("\n");    
  }


  /////////////////////////////////////////////////////////////////////////////////
  // RUN
  /////////////////////////////////////////////////////////////////////////////////

  // test settings
  if(0)
  {
    LsPar->DebugPrint = 7;
    // LsPar->DebugDraw = 3;
    // LsPar->PrintHostCommand = 15;
     
    // LsPar->SkipNoLaserPoints = 1;

    // NbrLines[0] = NbrLines[1] = LsPar->NbrLines = 0; 
    // DataMode[0] = DataMode[1] = LsPar->DataMode = 8; 
     
    // DataMode[0] = DataMode[1] = -1; // No TCP/IP Data Transfer after camera boot 

    // LsPar->Shutter = 200;
    LsPar->Laser = 2;
    
    // test ASCII
    // LsPar->CommandModeASCII = 1;
    // LsPar->DebugDraw = 3;
    // LsPar->PrintHostCommand = 7;
     
    // LsPar->Trigger = -1; 
    // LsPar->AutoTriggerFPS = 100;
    // LsPar->AutoTriggerError     =   0;
    // LsPar->EthernetPackNr = 1;
     LsPar->ExposureMode = 1;
    // LsPar->ROI_DY = 256;
    // LsPar->Algorithm = 0;
    // LsPar->PrintHostCommand = 3;
    // LsPar->HorLine = 0;
    // LsPar->DoubleShutter[0] = 200;
    // LsPar->DoubleShutter[1] = 4000;
    // LsPar->SubSample = 2;
  }

  /////////////////////////////////////////////////////////////////////////////////
  // start 3D Scanner
  /////////////////////////////////////////////////////////////////////////////////

  print("\n--- ");
  print("%s Version: Main=%d.%d.%d / Lib=%d", "VC 3D Laser Scanner", LsPar->MainVersion/10, LsPar->MainVersion%10, MAIN_HOT_FIX, LsPar->LibVersion);
  #ifdef PRODUCT_VERSION
    print(" / Product=%d", ProdPar->ProductVersion);
  #endif
  print(" ---\n\n");

  print("Press ENTER for hand operated scan\n");
  print("Press  ESC  to quit program\n");
  print("Press  'm'  to change parameter\n");
  print("Press  's'  to save BMP image for debug\n");
  print("Press  'r'  to read BMP image for debug\n");
  print("\n");

  
  /////////////////////////////////////////////////////////////////////////////////
  // reset sensor and set scanner parameter
  /////////////////////////////////////////////////////////////////////////////////

  // switch on laser for 500ms to show, sensor is ready
  SwitchLaser(1);
  time_delay(500);
  
  SensorReset(LsPar);


  /////////////////////////////////////////////////////////////////////////////////
  // clear output matrix
  /////////////////////////////////////////////////////////////////////////////////
  
  outPLC(0);


  /////////////////////////////////////////////////////////////////////////////////
  // scanning loop
  /////////////////////////////////////////////////////////////////////////////////

  do
  {
    //////////////////////////////////////////////
    // set start values
    //////////////////////////////////////////////
    
    // check for active Data Transfer Modes
    if( (DataMode[0] && NbrLines[0] && (NbrLines[0] != -2)) || (NbrLines[1] == -2) )
    {
      LsPar->NbrLines = NbrLines[0];
      LsPar->DataMode = DataMode[0];
    }
    else 
    {
      LsPar->NbrLines = NbrLines[1];
      LsPar->DataMode = DataMode[1];
    }
      
    // if all ports are inactiv, reset sensor settings  
    if ( (LsPar->NbrLines == 0) || (LsPar->DataMode == 0) )
    {
      // prepare sensor to wait for the first image
      ScreenBuffer2(LsPar, SCREEN_BUFF_FIRST);

      // reset auto trigger timer
      AutoTrigMode = 0;

      // reset double exposer counter
      LsPar->DoubleShutterImg = 0;
    }


    // Empty Send Buffer for inactive ports    
    for(PN=0; PN<2; PN++) 
    {
      if ( (NbrLines[PN] == 0) || (DataMode[PN] == 0) )
      {
        // send remaining TCP/IP data
        response = TCP_SendCommand(LsPar, alias[PN], LsPar->SendBuffer[PN], LsPar->MaxSendBuff, EMPTY_SEND_BUFFER);
      }
    }

    
    //////////////////////////////////////////////
    // look for PLC, Ethernet or telnet input
    //////////////////////////////////////////////

    // telnet input
    response = ReadKeyboard_Telnet(0);

    if (response == KEY_ESC  ) {RunMode = 0;}
    if (response == 'm'      ) {PN=0; ChangeParameter(LsPar); AutoTrigMode = 0;}
    if (response == 's'      ) {StoreImage();}
    if (response == 'r'      ) {ReadImage(); goto CALC_READ_IMAGE;}

    // if (response == '1'      ) {print("ON MODE\n");    LsPar->Laser = 1; SwitchLaser(LsPar->Laser);}
    // if (response == '2'      ) {print("PULSE MODE\n"); LsPar->Laser = 2; SwitchLaser(LsPar->Laser);}

    if (response == KEY_ENTER) 
    {
      // single shot 
      if(DataMode[0] && NbrLines[0]==0) {LsPar->NbrLines = NbrLines[0] = 1; LsPar->DataMode = DataMode[0];}
      if(DataMode[1] && NbrLines[1]==0) {LsPar->NbrLines = NbrLines[1] = 1; LsPar->DataMode = DataMode[1];}
    }

    // Ethernet input
    for(PN=0; PN<2; PN++) 
    {
      response = TCP_NewConnection(sock[PN], &alias[PN]);
      
      // <0 = error / 0 = new data / 1 = new connection
      if(response >= 0)
      {
        if(response == 1) 
        {
          // reset Counter
          TcpRecveCnt[PN] = 0;
          TcpSendCnt [PN] = 0;
        }

        // check for TCP/IP commands
        if(alias[PN]) response = TCP_ReadCommand(LsPar, &alias[PN], LsPar->ReceiveBuffer[PN], LsPar->MinReceiveBuff, LsPar->MaxReceiveBuff);

        if(LsPar->DebugPrint == 5) print("TCP_ReadCommand = %4d: PN=%d NbrLines=%d DataMode=%d\n", response, PN, NbrLines[PN], DataMode[PN]);
      }
    }

    //////////////////////////////////////////////
    // do image acquisition
    //////////////////////////////////////////////

    // print("NBR_LINES=%d DATA_MODE=%d [PN0:%d/%d  PN1:%d/%d]\n", LsPar->NbrLines, LsPar->DataMode, NbrLines[0], DataMode[0], NbrLines[1], DataMode[1]); time_delay(300);

    if( (LsPar->DataMode) && (LsPar->NbrLines) )
    {
      I32 ReadyImage  = SCREEN_RES_INI;

      // auto trigger active? (AUTO_TRIGGER_MODE [-1], AUTO_TRIGGER_POS after pos. edge [-2] or AUTO_TRIGGER_NEG after neg. edge [-3])
      if(LsPar->Trigger < 0)
      {
        static I32 TriggerStatus = -1;

        if(AutoTrigMode == 0)
        {
          ReadyImage = SCREEN_RES_BUSY;

          switch(LsPar->Trigger)
          {
            case  AUTO_TRIGGER_POS:
                  if(TRIGINP_PIN == 0                      ) TriggerStatus = 0;
                  if(TRIGINP_PIN != 0 && TriggerStatus == 0) TriggerStatus = 1;
                  break;

            case  AUTO_TRIGGER_NEG:
                  if(TRIGINP_PIN != 0                      ) TriggerStatus = 0;
                  if(TRIGINP_PIN == 0 && TriggerStatus == 0) TriggerStatus = 1;
                  break;

            default:
                  TriggerStatus = 1;
          }

          if(TriggerStatus == 1)
          {
            AutoTrigger(LsPar, AUTO_TRIG_INI);

            ReadyImage    = SCREEN_RES_INI;
            AutoTrigMode  =  1;
            TriggerStatus = -1;
          }
        }
        else
        {
          if( (LsPar->ExposureMode != DOUBLE_EXPOSURE) || (LsPar->DoubleShutterImg == 1) )
          {
            ReadyImage = SCREEN_RES_BUSY;

            if(AutoTrigger(LsPar, AUTO_TRIG_DIFF) == AUTO_TRIG_READY)
            {
              ReadyImage = ScreenBuffer2(LsPar, SCREEN_BUFF_SWAP);
              if(ReadyImage != SCREEN_RES_IMAGE) LsPar->AutoTriggerError++;
            }
          }
        }
      }

      // start next image acquisition if an image is ready
      if(ReadyImage == SCREEN_RES_INI)
      {
        ReadyImage = ScreenBuffer2(LsPar, SCREEN_BUFF_SWAP);
      }


      //////////////////////////////////////////////
      // calc image
      //////////////////////////////////////////////
      
      if (ReadyImage == SCREEN_RES_IMAGE)
      {
        CALC_READ_IMAGE:
      
        //////////////////////////////////////////////
        // detect laser line
        //////////////////////////////////////////////

        Ms1 = GetTimeMs();
        {
          image ImgArea;
          
          ImageAssign(&ImgArea, ScrByteAddr(0, 0), LsPar->ROI_DX, LsPar->ROI_DY_SubSample, ScrGetPitch);
          
          Error = ScannerDetection(LsPar, &ImgArea);
        }
        MsD = GetTimeMs() - Ms1;

        if(Error) print("ScannerDetection Error %d.\n", Error);
        
        //////////////////////////////////////////////
        // product calculation
        //////////////////////////////////////////////

        if(ReadyForCalc(LsPar))
        {
          // select all valid detections (if not, check if detection is inside ROI or ROI_MM)
          if(LsPar->SkipNoLaserPoints) DelNoneDetectionPoints(LsPar);
          
          // compensate sub sample
          if(LsPar->SubSample) 
          {
            I32 i, Fact, Nr = LsPar->LaserPoints;
            
            Fact = 1 << LsPar->SubSample;
            for(i=0; i<Nr; i++) LsPar->pLaserPosZ[i] *= Fact;
          }
          
          // convert from IMAGE to WORLD coordinates for measurements (mm)
          ImgToWorldCoord(LsPar, LsPar->LaserPoints, LsPar->pLaserPosX, LsPar->pLaserPosZ, LsPar->pLaserPosX_MM, LsPar->pLaserPosZ_MM);
          
          // calc product results
          #ifdef PRODUCT_VERSION
          Product_Measure(LsPar);
          #endif
          
          // calc auto shutter value (SkipNoLaserPoints should be active!)
          if(LsPar->ExposureMode == AUTO_EXPOSURE) AutoShutter(LsPar);         
          

          //////////////////////////////////////////////
          // send results via TCP/IP
          //////////////////////////////////////////////

          for(PN=0; PN<2; PN++) 
          {
            if(NbrLines[PN]) response = TCP_SendCommand(LsPar, alias[PN], LsPar->SendBuffer[PN], LsPar->MaxSendBuff, USE_SEND_BUFFER);
          }
          
          if( (NbrLines[0] > 0) && (DataMode[0]) ) NbrLines[0]--;
          if( (NbrLines[1] > 0) && (DataMode[1]) ) NbrLines[1]--;
        }

        if(LsPar->DebugPrint == 1) print("SysMemFree=%6d\n", sysmemfree());
        if(LsPar->DebugPrint == 2) {static I32 cnt=0; if(++cnt == 50) {print("t=%2dms fps=%5.1f  Error=%d\n", (I32)MsD, (float)(cnt*1000)/(float)(Ms1-Ms0), Error); Ms0=Ms1; cnt=0;}}
      }
    }
  }
  while(RunMode);

  /////////////////////////////////////////////////////////////////////////////////
  // Stop RUN mode
  /////////////////////////////////////////////////////////////////////////////////

  QUIT:

  for(PN=0; PN<2; PN++) 
  {
    if(alias[PN]) TCP_Close(alias[PN]);
    if(sock [PN]) TCP_Close(sock [PN]);
  }
  
  ScreenBuffer2(LsPar, SCREEN_BUFF_DEINI);

  #ifdef PRODUCT_VERSION
    ProductParameterIni(LsPar, MEM_DEINI);
  #endif

  SensorParameterIni(LsPar, MEM_DEINI);
  ScannerInit(LsPar, MEM_DEINI);

  vmode(vmLive);
  

  /////////////////////////////////////////////////////////////////////////////////
  // set lower task priority
  /////////////////////////////////////////////////////////////////////////////////

  task_set_priority(id, oldprio, &highprio); // reset Taskpriority (old value)
  print("Set own task '%s' to prio %d! (highprio=%d)\n", task_get_name_from_id(id), oldprio, highprio);
}

/*************************************************************************************************************************/

I32 SensorParameterIni (LsParameter *LsPar, I32 Mode)
{
  I32        Error=0;
  static I32 MemoryReady=0;

  /////////////////////////////////////////////////////////////////////////////////
  // release memory (DeIni or NewAllocation)
  /////////////////////////////////////////////////////////////////////////////////

  if(Mode == MEM_DEINI || Mode == MEM_INI)
  {
    if(MemoryReady) 
    {
      MemoryReady = 0;
      
      if(LsPar->SendBuffer[1]    != NULL) sysfree(LsPar->SendBuffer[1]);
      if(LsPar->SendBuffer[0]    != NULL) sysfree(LsPar->SendBuffer[0]);
      if(LsPar->ReceiveBuffer[1] != NULL) sysfree(LsPar->ReceiveBuffer[1]);
      if(LsPar->ReceiveBuffer[0] != NULL) sysfree(LsPar->ReceiveBuffer[0]);
      if(LsPar->pSensorSetting   != NULL) sysfree(LsPar->pSensorSetting);
    }
  }


  /////////////////////////////////////////////////////////////////////////////////
  // allocate memory 
  /////////////////////////////////////////////////////////////////////////////////

  if( Mode == MEM_INI || (MemoryReady == 0 && Mode != MEM_DEINI) )
  {
    Error       = 0;
    MemoryReady = 1;
    
    // ini memory pointer
    LsPar->pSensorSetting   = NULL;
    LsPar->ReceiveBuffer[0] = NULL;
    LsPar->ReceiveBuffer[1] = NULL;
    LsPar->SendBuffer[0]    = NULL;
    LsPar->SendBuffer[1]    = NULL;

    // memory for sensor setting parameter
    LsPar->SensorSettingSize = 0;
    LsPar->pSensorSetting    = (I8 *) sysmalloc( (MAX_SETTING_SIZE * sizeof(  U8) - 1) / sizeof(I32) + 1, MDATA);
    if(LsPar->pSensorSetting == NULL) {Error = ERR_MEM_INI;}  

    // memory for receive buffer 0
    LsPar->MaxReceiveBuff    = GLOBAL_HEAD_SIZE + MAX_SETTING_SIZE;
    LsPar->ReceiveBuffer[0]  = (U8  *) sysmalloc( ((LsPar->MaxReceiveBuff) * sizeof(   U8) - 1) / sizeof(I32) + 1, MDATA);
    if(LsPar->ReceiveBuffer[0] == NULL) {Error = ERR_MEM_INI;}

    // memory for receive buffer 1
    LsPar->MaxReceiveBuff    = GLOBAL_HEAD_SIZE + MAX_SETTING_SIZE;
    LsPar->ReceiveBuffer[1]  = (U8  *) sysmalloc( ((LsPar->MaxReceiveBuff) * sizeof(   U8) - 1) / sizeof(I32) + 1, MDATA);
    if(LsPar->ReceiveBuffer[1] == NULL) {Error = ERR_MEM_INI;}

    // memory for send buffer 0 [(ScrGetColumns * ScrGetRows + GLOBAL_HEAD_SIZE) .. (sysmemfree() - 8 * ScrGetColumns * ScrGetRows)]
    LsPar->MaxSendBuff       = 512 * GLOBAL_HEAD_SIZE + 4 * ScrGetColumns * ScrGetRows; // memory for 512 scans or 4 images
    LsPar->SendBuffer[0]     = (U8 *) sysmalloc( (LsPar->MaxSendBuff * sizeof(U8) - 1) / sizeof(I32) + 1, MDATA);
    if(LsPar->SendBuffer[0]  == NULL) {Error = ERR_MEM_INI;}
    
    // memory for send buffer 1 [(ScrGetColumns * ScrGetRows + GLOBAL_HEAD_SIZE) .. (sysmemfree() - 8 * ScrGetColumns * ScrGetRows)]
    LsPar->MaxSendBuff       = 512 * GLOBAL_HEAD_SIZE + 4 * ScrGetColumns * ScrGetRows; // memory for 512 scans or 4 images
    LsPar->SendBuffer[1]     = (U8 *) sysmalloc( (LsPar->MaxSendBuff * sizeof(U8) - 1) / sizeof(I32) + 1, MDATA);
    if(LsPar->SendBuffer[1]  == NULL) {Error = ERR_MEM_INI;}
    
    if(Error) 
    {
      print("Error: Memory error in SensorParameterIni()\n");
      goto QUIT_INI;
    }
  }


  /////////////////////////////////////////////////////////////////////////////////
  // ini sensor parameter
  /////////////////////////////////////////////////////////////////////////////////

  if( Mode == MEM_INI || (MemoryReady == 0 && Mode != MEM_DEINI) )
  {
    // for debugging print host commands and drawings
    LsPar->PrintHostCommand  = 0;
    LsPar->DebugPrint        = 0;
    LsPar->DebugDraw         = 0;

    // time stamp for image exposer ready time in ms (since camera boot)
    LsPar->ImgTimeStampMs    = 0;   

    // set scanner default parameter
    LsPar->StepDx            = 1;

    // double exposure
    LsPar->DoubleShutterImg  = 0;

    // program version
    LsPar->MainVersion       = MAIN_VERSION;

    // read camera SNR
    LsPar->CameraSNR         = GetCameraID();
  
    // set scanner default parameter
    LsPar->Stop              = 0; 

    // set LaserPoints
    LsPar->LaserPoints       = 0;    

    // reset autotrigger error    
    LsPar->AutoTriggerError  = 0;
    
    // subsample parameters
    LsPar->ROI_Y_SubSample   = 0;
    LsPar->ROI_DY_SubSample  = 0;
    
    // PLC parameters
    LsPar->InPLC             = 0;
    LsPar->OutPLC            = 0;    
  
    // ini product setting string
    LsPar->SensorSettingSize = 0;
    LsPar->pSensorSetting[0] = 0;

    // Load sensor setting from file. Use default values for any missing parameters.
    Error = SensorParameterExchange(LsPar, PARAMETER_LOAD);
  }


  /////////////////////////////////////////////////////////////////////////////////
  // ready
  /////////////////////////////////////////////////////////////////////////////////  

  QUIT_INI:
  
  LsPar->Error = Error;
  
  return Error;
}

/*************************************************************************************************************************/

I32 SensorParameterRange (LsParameter *LsPar, void *pPar, I32 *SetVal, I32 *MinVal, I32 *MaxVal)
{
  I32 Error = 0;
  F32 MinX, MaxX, MinZ, MaxZ;
    
    

  /////////////////////////////////////////////////////////////////////////////////
  // ini values
  /////////////////////////////////////////////////////////////////////////////////

  *MinVal = 0;
  *MaxVal = 0;
  *SetVal = 0;

  // calc max range in world coordinates  
  GetMaxSize(LsPar, 0, &MinX, &MaxX, &MinZ, &MaxZ);
  

  /////////////////////////////////////////////////////////////////////////////////
  // parameter range
  /////////////////////////////////////////////////////////////////////////////////

  if (pPar == &LsPar->Stop) 
  {
    *MinVal = 0;
    *MaxVal = 0;
    *SetVal = 0;
  }

  if (pPar == &LsPar->Laser) 
  {
    *MinVal = 0;
    *MaxVal = 2;
    *SetVal = 0;
  }

  if (pPar == &LsPar->ROI_X) 
  {
    *MinVal = 0;
    *MaxVal = ScrGetColumns - 3;
    *SetVal = 0;
  }

  if (pPar == &LsPar->ROI_Y) 
  {
    *MinVal = 0;
    *MaxVal = ScrGetRows - 3;
    *SetVal = 0;
  }

  if (pPar == &LsPar->ROI_DX)
  {
    *MinVal = 6;
    *MaxVal = ScrGetColumns;
    *SetVal = ScrGetColumns;
  }

  if (pPar == &LsPar->ROI_DY) 
  {
    *MinVal = 6;
    *MaxVal = ScrGetRows;
    *SetVal = ScrGetRows;
  }

  if (pPar == &LsPar->GainVal)
  {
    *MinVal = 0;
    *MaxVal = 1023;
    *SetVal = getvar(GAIN);
  }

  if (pPar == &LsPar->Shutter) 
  {
    *MinVal = 5;
    *MaxVal = 500000;
    *SetVal = 500;
  }

  if (pPar == &LsPar->Trigger) 
  {
    *MinVal = -3;
    *MaxVal =  2;
    *SetVal =  0;
  }

  if (pPar == &LsPar->DataMode) 
  {
    *MinVal = -1;
    *MaxVal =  9;
    *SetVal =  0;
  }

  if (pPar == &LsPar->NbrLines) 
  {
    *MinVal = -2;
    *MaxVal = 10000;
    *SetVal = 0;
  }

  if (pPar == &LsPar->Methode) 
  { 
    *MinVal = 0;
    *MaxVal = 2;
    *SetVal = 2;
  }

  if (pPar == &LsPar->Algorithm) 
  {
    *MinVal = 0;
    *MaxVal = 1;
    *SetVal = 0;
  }

  if (pPar == &LsPar->RlcThresh) 
  {
    *MinVal =   1;
    *MaxVal = 254;
    *SetVal =  48;
  }

  if (pPar == &LsPar->MinWidth)
  {
    *MinVal =   1;
    *MaxVal = 100;
    *SetVal =   1;
  }

  if (pPar == &LsPar->MaxWidth)
  {
    *MinVal =   1;
    *MaxVal = 100;
    *SetVal =  30;
  }

  if (pPar == &LsPar->ExposureMode) 
  {
    *MinVal = 0;
    *MaxVal = 3;
    *SetVal = FIX_EXPOSURE;
  }

  if (pPar == &LsPar->AutoShutterVal) 
  { 
    *MinVal = 1;
    *MaxVal = 254;
    *SetVal = 128;
  }
  
  if (pPar == &LsPar->AutoShutterFilter) 
  { 
    *MinVal = 0;
    *MaxVal = 6;
    *SetVal = 0;
  }
  

  if (pPar == &LsPar->AutoShutterMin) 
  {
    *MinVal = 5;
    *MaxVal = 100000;
    *SetVal = 10;
  }

  if (pPar == &LsPar->AutoShutterMax) 
  {
    *MinVal = 5;
    *MaxVal = 100000;
    *SetVal = 1000;
  }

  if (pPar == &LsPar->DoubleShutter[0]) 
  {
    *MinVal = 5;
    *MaxVal = 100000;
    *SetVal = 100;
  }

  if (pPar == &LsPar->DoubleShutter[1])
  {
    *MinVal = 5;
    *MaxVal = 100000;
    *SetVal = 500;
  }

  if (pPar == &LsPar->ReflexionFilter) 
  {
    *MinVal = -8;
    *MaxVal =  8;
    *SetVal =  0;
  }
  
  if (pPar == &LsPar->LaserMaskFilter) 
  {
    *MinVal =   0;
    *MaxVal = 255;
    *SetVal =   0;
  }

  if (pPar == &LsPar->ROI_X_MM) 
  {
    *MinVal = Round(MinX);
    *MaxVal = Round(MaxX);
    *SetVal = Round(MinX);
  }

  if (pPar == &LsPar->ROI_Z_MM) 
  {
    *MinVal = Round(MinZ);
    *MaxVal = Round(MaxZ);
    *SetVal = Round(MinZ);
  }

  if (pPar == &LsPar->ROI_DX_MM) 
  {
    *MinVal = 1;
    *MaxVal = Round(MaxX - MinX);
    *SetVal = Round(MaxX - MinX);
  }

  if (pPar == &LsPar->ROI_DZ_MM) 
  {
    *MinVal = 1;
    *MaxVal = Round(MaxZ - MinZ);
    *SetVal = Round(MaxZ - MinZ);
  }

  if (pPar == &LsPar->LibVersion) 
  { 
    *MinVal = LsPar->LibVersion;
    *MaxVal = LsPar->LibVersion;
    *SetVal = LsPar->LibVersion;
  }

  if (pPar == &LsPar->EthernetPackNr) 
  {
    *MinVal = 0;
    *MaxVal = 100;
    *SetVal = 1;
  }

  if (pPar == &LsPar->AutoTriggerFPS) 
  { 
    *MinVal = 1;
    *MaxVal = 10000;
    *SetVal = 50 * 10; // 1/10 Hz
  }

  if (pPar == &LsPar->AutoTriggerError) 
  { 
    *MinVal = 0;
    *MaxVal = 2147483647; // Max I32
    *SetVal = 0;
  }

  if (pPar == &LsPar->EthernetSendNoWait) 
  {
    *MinVal = 0;
    *MaxVal = 1;
    *SetVal = 1;
  }

  if (pPar == &LsPar->SubSample)
  {
    *MinVal = 0;
    *MaxVal = 2;
    *SetVal = 0;
  }

  if (pPar == &LsPar->CameraSNR) 
  { 
    *MinVal = GetCameraID();
    *MaxVal = GetCameraID();
    *SetVal = GetCameraID();
  }

  if (pPar == &LsPar->MedianFilter) 
  { 
    *MinVal = 0;
    *MaxVal = 4;
    *SetVal = 0;
  }

  if (pPar == &LsPar->SpeckleFilter) 
  { 
    *MinVal = 0;
    *MaxVal = 4;
    *SetVal = 3;
  }

  if (pPar == &LsPar->SkipNoLaserPoints) 
  { 
    *MinVal = 0;
    *MaxVal = 1;
    *SetVal = 1;
  }

  if (pPar == &LsPar->InPLC) 
  { 
    *MinVal = 0;
    *MaxVal = 15;
    *SetVal = 0;
  }

  if (pPar == &LsPar->OutPLC) 
  { 
    *MinVal = 0;
    *MaxVal = 15;
    *SetVal = 0;
  }
  
  /*
  if (pPar == &LsPar->CommandModeASCII) 
  { 
    *MinVal = 0;
    *MaxVal = 1;
    *SetVal = 0;
  }
  */
  
  if (pPar == &LsPar->StepDx) 
  { 
    *MinVal = 1;
    *MaxVal = 8;
    *SetVal = 1;
  }
  
  /////////////////////////////////////////////////////////////////////////////////
  // check value
  /////////////////////////////////////////////////////////////////////////////////

  if(*SetVal < *MinVal) {*SetVal = *MinVal; Error = ERR_PARM_RANGE_MIN;}
  if(*SetVal > *MaxVal) {*SetVal = *MaxVal; Error = ERR_PARM_RANGE_MAX;}
  
  
  return Error;
}

/*************************************************************************************************************************/

I32 SensorParameterExchange (LsParameter *LsPar, enum ParMode Mode)
{
  I32      i, Error, Result, Val, SetVal, MinVal, MaxVal;
  I32      Debug, ExchangeMode, SettingPos, MissingParameter;
  I32      RoiMMSet=0, RoiPixSet=0;
  I32      *pParStruct;
  I8       *pParName, SettingString[200];  
  I8       *PAR_NAME[50];
  I32      *PAR_ADDR[50], ParCnt; 
  enum     ExchangeMode {GET_VAL = 0, SET_VAL, INI_VAL};

  
  /////////////////////////////////////////////////////////////////////////////////
  // ini value
  /////////////////////////////////////////////////////////////////////////////////
  
  Error            = 0;
  SettingPos       = 0;
  Debug            = 0;
  ParCnt           = 0;
  MissingParameter = 0;
  ExchangeMode     = GET_VAL;

  if(LsPar->DebugPrint == 8) Debug = 1; 


  /////////////////////////////////////////////////////////////////////////////////
  // generate ASCII parameter name and corresponding struct address table
  /////////////////////////////////////////////////////////////////////////////////
  
  PAR_NAME[ParCnt] = "LaserMode";            PAR_ADDR[ParCnt] = &LsPar->Laser;                ParCnt++;
  PAR_NAME[ParCnt] = "ROI_X_MM";             PAR_ADDR[ParCnt] = &LsPar->ROI_X_MM;             ParCnt++;
  PAR_NAME[ParCnt] = "ROI_Z_MM";             PAR_ADDR[ParCnt] = &LsPar->ROI_Z_MM;             ParCnt++;
  PAR_NAME[ParCnt] = "ROI_DX_MM";            PAR_ADDR[ParCnt] = &LsPar->ROI_DX_MM;            ParCnt++;
  PAR_NAME[ParCnt] = "ROI_DZ_MM";            PAR_ADDR[ParCnt] = &LsPar->ROI_DZ_MM;            ParCnt++;
  PAR_NAME[ParCnt] = "GainVal";              PAR_ADDR[ParCnt] = &LsPar->GainVal;              ParCnt++;
  PAR_NAME[ParCnt] = "ShutterTime";          PAR_ADDR[ParCnt] = &LsPar->Shutter;              ParCnt++;
  PAR_NAME[ParCnt] = "TriggerMode";          PAR_ADDR[ParCnt] = &LsPar->Trigger;              ParCnt++;
  PAR_NAME[ParCnt] = "DataMode";             PAR_ADDR[ParCnt] = &LsPar->DataMode;             ParCnt++;
  PAR_NAME[ParCnt] = "NbrLines";             PAR_ADDR[ParCnt] = &LsPar->NbrLines;             ParCnt++;
  PAR_NAME[ParCnt] = "Methode";              PAR_ADDR[ParCnt] = &LsPar->Methode;              ParCnt++;
  PAR_NAME[ParCnt] = "Algorithm";            PAR_ADDR[ParCnt] = &LsPar->Algorithm;            ParCnt++;
  PAR_NAME[ParCnt] = "RlcThresh";            PAR_ADDR[ParCnt] = &LsPar->RlcThresh;            ParCnt++;
  PAR_NAME[ParCnt] = "MinWidth";             PAR_ADDR[ParCnt] = &LsPar->MinWidth;             ParCnt++;
  PAR_NAME[ParCnt] = "MaxWidth";             PAR_ADDR[ParCnt] = &LsPar->MaxWidth;             ParCnt++;
  PAR_NAME[ParCnt] = "ExposureMode";         PAR_ADDR[ParCnt] = &LsPar->ExposureMode;         ParCnt++;
  PAR_NAME[ParCnt] = "AutoShutterVal";       PAR_ADDR[ParCnt] = &LsPar->AutoShutterVal;       ParCnt++;
  PAR_NAME[ParCnt] = "AutoShutterFilter";    PAR_ADDR[ParCnt] = &LsPar->AutoShutterFilter;    ParCnt++;
  PAR_NAME[ParCnt] = "AutoShutterMin";       PAR_ADDR[ParCnt] = &LsPar->AutoShutterMin;       ParCnt++;
  PAR_NAME[ParCnt] = "AutoShutterMax";       PAR_ADDR[ParCnt] = &LsPar->AutoShutterMax;       ParCnt++;
  PAR_NAME[ParCnt] = "DoubleShutter1";       PAR_ADDR[ParCnt] = &LsPar->DoubleShutter[0];     ParCnt++;
  PAR_NAME[ParCnt] = "DoubleShutter2";       PAR_ADDR[ParCnt] = &LsPar->DoubleShutter[1];     ParCnt++;
  PAR_NAME[ParCnt] = "ReflexionFilter";      PAR_ADDR[ParCnt] = &LsPar->ReflexionFilter;      ParCnt++;
  PAR_NAME[ParCnt] = "EthernetPackNr";       PAR_ADDR[ParCnt] = &LsPar->EthernetPackNr;       ParCnt++;
  PAR_NAME[ParCnt] = "AutoTriggerFPS";       PAR_ADDR[ParCnt] = &LsPar->AutoTriggerFPS;       ParCnt++;
  PAR_NAME[ParCnt] = "AutoTriggerError";     PAR_ADDR[ParCnt] = &LsPar->AutoTriggerError;     ParCnt++;
  PAR_NAME[ParCnt] = "EthernetSendNoWait";   PAR_ADDR[ParCnt] = &LsPar->EthernetSendNoWait;   ParCnt++;
  PAR_NAME[ParCnt] = "SubSample";            PAR_ADDR[ParCnt] = &LsPar->SubSample;            ParCnt++;
  PAR_NAME[ParCnt] = "MedianFilter";         PAR_ADDR[ParCnt] = &LsPar->MedianFilter;         ParCnt++;
  PAR_NAME[ParCnt] = "SpeckleFilter";        PAR_ADDR[ParCnt] = &LsPar->SpeckleFilter;        ParCnt++;
  PAR_NAME[ParCnt] = "SkipNoLaserPoints";    PAR_ADDR[ParCnt] = &LsPar->SkipNoLaserPoints;    ParCnt++;
  PAR_NAME[ParCnt] = "CameraSNR";            PAR_ADDR[ParCnt] = &LsPar->CameraSNR;            ParCnt++;
  PAR_NAME[ParCnt] = "LibVersion";           PAR_ADDR[ParCnt] = &LsPar->LibVersion;           ParCnt++;
  PAR_NAME[ParCnt] = "LaserMaskFilter";      PAR_ADDR[ParCnt] = &LsPar->LaserMaskFilter;      ParCnt++;
  
  // SmartShape needs ROI in Pix
  PAR_NAME[ParCnt] = "ROI_X_PIX";            PAR_ADDR[ParCnt] = &LsPar->ROI_X;                ParCnt++;
  PAR_NAME[ParCnt] = "ROI_Y_PIX";            PAR_ADDR[ParCnt] = &LsPar->ROI_Y;                ParCnt++;
  PAR_NAME[ParCnt] = "ROI_DX_PIX";           PAR_ADDR[ParCnt] = &LsPar->ROI_DX;               ParCnt++;
  PAR_NAME[ParCnt] = "ROI_DY_PIX";           PAR_ADDR[ParCnt] = &LsPar->ROI_DY;               ParCnt++;
  
  
  /////////////////////////////////////////////////////////////////////////////////
  // MODE - Pre Settings
  /////////////////////////////////////////////////////////////////////////////////

  // read file settings and set all missing parameters to standard value
  if(Mode == PARAMETER_LOAD)
  {
    ExchangeMode = INI_VAL;

    // read parameter file to RAM    
    Result = ReadFileToMem(SENSOR_FILE, LsPar->pSensorSetting, MAX_SETTING_SIZE, Debug); 

    if(Result >= 0)
    {
      LsPar->SensorSettingSize = Result;
      LsPar->SensorSettingSize = DeleteComment(LsPar->pSensorSetting, LsPar->SensorSettingSize, LsPar->pSensorSetting);

      if(getvar(TELNET)) print("Reading sensor parameter file %s successful.\n", SENSOR_FILE);
    }
    else
    {
      LsPar->SensorSettingSize = 0;
      LsPar->pSensorSetting[0] = 0;
      
      if(getvar(TELNET)) print("Sensor parameter file %s doesn't exist.\n", SENSOR_FILE);
    }    
  }

 
  // save parameter to file
  if(Mode == PARAMETER_SAVE)
  {
    ExchangeMode = GET_VAL;

    // for auto run after sensor booting
    // LsPar->DataMode  = DataMode[0];
    // LsPar->NbrLines  = NbrLines[0];
    
    // if(LsPar->DataMode == 0) LsPar->DataMode = -1; 
    // if(LsPar->NbrLines == 0) LsPar->NbrLines = -1;
  }
  
  
  // read sensor settings parameter and ignore all missing parameters
  if(Mode == PARAMETER_RECEIVE)
  {
    ExchangeMode = SET_VAL;
    LsPar->SensorSettingSize = DeleteComment(LsPar->pSensorSetting, LsPar->SensorSettingSize, LsPar->pSensorSetting);

    // print("Mode = PARAMETER_RECEIVE: %s\n", LsPar->pSensorSetting);
  }
  
  
  // send file parameter settings
  if(Mode == PARAMETER_SEND)
  {
    ExchangeMode = GET_VAL;
  }
  
  
  
  /////////////////////////////////////////////////////////////////////////////////
  // read / write parameter settings structure / pSensorSetting
  /////////////////////////////////////////////////////////////////////////////////
  
  if(Error == 0)
  {
    for(i=0; i < ParCnt; i++)
    {
      pParName   = PAR_NAME[i];
      pParStruct = PAR_ADDR[i];

      // read parameter value and range
      SensorParameterRange(LsPar, pParStruct, &SetVal, &MinVal, &MaxVal);
      
      if(ExchangeMode == GET_VAL) 
      {
        // sprint(SettingString, "%s=%d\n", pParName, *pParStruct);
        sprint(SettingString, "%s=%d Min=%d Max=%d\n", pParName, *pParStruct, MinVal, MaxVal);

        strcpy(&LsPar->pSensorSetting[SettingPos], SettingString);
        
        SettingPos += strlen(SettingString);
      }
      else
      {
        Result = ReadIntValue(LsPar->pSensorSetting, pParName, &Val, Debug);
        
         // print("Result=%4d: %s=%d [%d..%d]  Adr=0x%x\n", Result, pParName, Val, MinVal, MaxVal, pParStruct);
        
        if(Result == 0)
        {
          if(Val < MinVal) {Val = MinVal; Error = ERR_PARM_RANGE_MIN;}
          if(Val > MaxVal) {Val = MaxVal; Error = ERR_PARM_RANGE_MAX;}
          
          *pParStruct = Val;
          
          // ROI settings in MM or Pixel?
          if(pParStruct == &LsPar->ROI_X    ) RoiPixSet++;
          if(pParStruct == &LsPar->ROI_Y    ) RoiPixSet++;
          if(pParStruct == &LsPar->ROI_DX   ) RoiPixSet++;
          if(pParStruct == &LsPar->ROI_DY   ) RoiPixSet++;
          
          if(pParStruct == &LsPar->ROI_X_MM ) RoiMMSet++;
          if(pParStruct == &LsPar->ROI_Z_MM ) RoiMMSet++;
          if(pParStruct == &LsPar->ROI_DX_MM) RoiMMSet++;
          if(pParStruct == &LsPar->ROI_DZ_MM) RoiMMSet++;

          // set DataMode and NbrLines
          if(PN > 1) PN = 0;
          if(pParStruct == &LsPar->NbrLines ) NbrLines[PN] = LsPar->NbrLines;
          if(pParStruct == &LsPar->DataMode ) DataMode[PN] = LsPar->DataMode;
        }
        else
        {
          if(ExchangeMode == INI_VAL) *pParStruct = SetVal;
          
          MissingParameter++;
        }
      }
    }
    
    // if we don't find any fitting command 
    if(MissingParameter == ParCnt) Error = ERR_PARM_NAME;
    
    // set setting size (+1 including termination byte 0, which is already there)
    LsPar->SensorSettingSize = strlen(LsPar->pSensorSetting) + 1;
  }
  
  
  /////////////////////////////////////////////////////////////////////////////////
  // MODE - Post Settings
  /////////////////////////////////////////////////////////////////////////////////

  if(Mode == PARAMETER_LOAD)
  {
    if(MissingParameter) if(getvar(TELNET)) print("The missing %d Sensor Parameter(s) will be replaced by default values.\n", MissingParameter);
  }
  
  if(Mode == PARAMETER_LOAD || Mode == PARAMETER_RECEIVE)
  {
    // set ROI in Pixel or MM
    // printf("SensorParameterExchange:\n");
    if(RoiMMSet || RoiPixSet)
    {
      if(RoiMMSet >= RoiPixSet) WorldToImg_ROI(LsPar);
      else                      ImgToWorld_ROI(LsPar);
    }
    
    // adjust sensor ROI according to JOB order
    if(LsPar->DataMode == 8 || LsPar->DataMode == 9)
    {
      #ifdef PRODUCT_VERSION  
      AdjustProductParameter(LsPar);
      #endif
    }

    // double exposure
    LsPar->DoubleShutterImg  = 0;

    // set scanner default parameter
    LsPar->Stop              = 0;  
    
    // ini real shutter value
    LsPar->ShutterReal       = LsPar->Shutter;

    // reset autotrigger error    
    LsPar->AutoTriggerError  = 0;

    // Global Header Size 
    LsPar->MinReceiveBuff       = GLOBAL_HEAD_SIZE; 
  }

  // save parameter to file
  if(Mode == PARAMETER_SAVE)
  {
    if(getvar(TELNET)) print("Store file %s. Please wait...\n", SENSOR_FILE);
    
    if(Debug) if(getvar(TELNET)) print("Sensor Parameter:\n-----------------\n%s\n", LsPar->pSensorSetting);

    Result = WriteMemToFile(SENSOR_FILE, LsPar->pSensorSetting, LsPar->SensorSettingSize, Debug);
    
    if(Result < 0) {Error = Result; if(getvar(TELNET)) print("Error %d.\n", Error);}
    else           {                if(getvar(TELNET)) print("Ready.\n");}
  }

  // save parameter to file
  if(Mode == PARAMETER_SHOW)
  {
    if(getvar(TELNET)) print("Sensor Parameter:\n-----------------\n%s\n", LsPar->pSensorSetting);
  }
  
  
  return Error;
}

/*************************************************************************************************************************/

void SensorParameterChange (LsParameter *LsPar, I32 *ParValue, I32 Step, I8 *text)
{
  I32 SetVal, MinVal, MaxVal;
  
  // get range
  SensorParameterRange(LsPar, ParValue, &SetVal, &MinVal, &MaxVal);
  
  // change value
  ChangeValue_Telnet(ParValue, MinVal, MaxVal, Step, text);
}

/*************************************************************************************************************************/

void ChangeParameter (LsParameter *LsPar)

{
  I32 response;

  EmptyKeyboardBuffer();

  do
  {
    /********************************************/
    /* open menu                                */
    /********************************************/

    print("\n");
    print("--------------------------------------------------------------------------------\n");
    print("Setup Parameter\n");
    print("--------------------------------------------------------------------------------\n");
    print("p:   set standard parameter\n");
    print("e:   set expert parameter\n");
    print("r:   reset all counters\n");
    print("s:   save sensor / product parameter\n");
    print("l:   load sensor / product parameter\n");
    print("d:   debug parameters\n");
    print("w:   show parmeter at telnet port\n");
    #ifdef PRODUCT_VERSION
    print("u:   set product parameter\n");
    #endif
    print("--------------------------------------------------------------------------------\n");
    print("q:   quit Menu\n");
    print("\n");


    response = ReadKeyboard_Telnet(1);

    /********************************************/
    /* menu                                     */
    /********************************************/

    switch(response)
    {
      case 'p' :
                 SensorParameterChange(LsPar, &LsPar->Laser,              1, "Set Laser: 0=OFF 1=ON 2=PULSE");
                 SensorParameterChange(LsPar, &LsPar->Trigger,            1, "Set Trigger: 0=OFF 1=PosEdge 2=NegEdge -1=AutoTrigger -2=AutoTrigPos -3=AutoTrigNeg");
                 if(LsPar->Trigger < 0)
                 {
                 SensorParameterChange(LsPar, &LsPar->AutoTriggerFPS,    10, "Set AutoTrigger Frames Per Second [1/10 Hz]");
                 SensorParameterChange(LsPar, &LsPar->AutoTriggerError,   1, "Set AutoTrigger Error");
                 }
                 SensorParameterChange(LsPar, &LsPar->ROI_X,              8, "Set ROI Area X [Pixel]");
                 SensorParameterChange(LsPar, &LsPar->ROI_Y,              8, "Set ROI Area Y [Pixel]");
                 SensorParameterChange(LsPar, &LsPar->ROI_DX,             8, "Set ROI Area DX [Pixel]");
                 SensorParameterChange(LsPar, &LsPar->ROI_DY,             8, "Set ROI Area DY [Pixel]");
                 
                 LsPar->NbrLines = NbrLines[0];
                 SensorParameterChange(LsPar, &LsPar->NbrLines,           1, "Set Nr_Lines at Port 0: -1=endless X=NrImages -2=FromPort1"); 
                 NbrLines[0] = LsPar->NbrLines; 
                 LsPar->NbrLines = NbrLines[1];
                 SensorParameterChange(LsPar, &LsPar->NbrLines,           1, "Set Nr_Lines at Port 1: -1=endless X=NrImages -2=FromPort0"); 
                 NbrLines[1] = LsPar->NbrLines;

                 LsPar->DataMode = DataMode[0];
                 SensorParameterChange(LsPar, &LsPar->DataMode,           1, "Set Data Mode at Port 0"); 
                 DataMode[0] = LsPar->DataMode;
                 LsPar->DataMode = DataMode[1];
                 SensorParameterChange(LsPar, &LsPar->DataMode,           1, "Set Data Mode at Port 1"); 
                 DataMode[1] = LsPar->DataMode;

                 SensorParameterChange(LsPar, &LsPar->ExposureMode,       1, "Set Exposure Mode: 0=FIX_EXPOSURE 1=AUTO_EXPOSURE 2=DOUBLE_EXPOSURE");

                 if(LsPar->ExposureMode == FIX_EXPOSURE)
                 {
                 SensorParameterChange(LsPar, &LsPar->Shutter,         1000, "Set Shutter [us] - 1 ms steps");
                 SensorParameterChange(LsPar, &LsPar->Shutter,            5, "Set Shutter [us]");
                 }

                 if(LsPar->ExposureMode == AUTO_EXPOSURE)
                 {
                 SensorParameterChange(LsPar, &LsPar->AutoShutterVal,     1, "Auto shutter Laser brightness");
                 SensorParameterChange(LsPar, &LsPar->AutoShutterFilter,  1, "Auto shutter filter (0:all 1:median 2:bright 3:dark 4:center 5:left 6:right detection points");
                 SensorParameterChange(LsPar, &LsPar->AutoShutterMin,     5, "Auto shutter time min [us]");
                 SensorParameterChange(LsPar, &LsPar->AutoShutterMax,    50, "Auto shutter time max [us]");
                 }

                 if(LsPar->ExposureMode == DOUBLE_EXPOSURE)
                 {
                 SensorParameterChange(LsPar, &LsPar->DoubleShutter[0],   1, "Double Exposure (Shutter 1 [us])");
                 SensorParameterChange(LsPar, &LsPar->DoubleShutter[1],   1, "Double Exposure (Shutter 2 [us])");
                 }
                 break;

      case 'e' :
                 SensorParameterChange(LsPar, &LsPar->GainVal,            1, "Set GAINVAL");
                 SensorParameterChange(LsPar, &LsPar->Methode,            1, "Set Methode");
                 SensorParameterChange(LsPar, &LsPar->Algorithm,          1, "Set Algorithm (0=I16 / 1=F32)");
                 SensorParameterChange(LsPar, &LsPar->RlcThresh,          1, "RLC Threshold");
                 SensorParameterChange(LsPar, &LsPar->SubSample,          1, "Image Sub Sample");
                 SensorParameterChange(LsPar, &LsPar->StepDx,             1, "Step Dx [Pixel]");
                 SensorParameterChange(LsPar, &LsPar->MinWidth,           1, "Min. Laser Width [Pixel]");
                 SensorParameterChange(LsPar, &LsPar->MaxWidth,           1, "Max. Laser Width [Pixel]");
                 SensorParameterChange(LsPar, &LsPar->EthernetPackNr,     1, "Ethernet Line Package Nr");
                 SensorParameterChange(LsPar, &LsPar->EthernetSendNoWait, 1, "Ethernet Send No Wait Mode");
                 SensorParameterChange(LsPar, &LsPar->ReflexionFilter,    1, "Laser Detection Filter");
                 SensorParameterChange(LsPar, &LsPar->MedianFilter,       1, "Median Filter: 0=no filter 1=3x1 2=5x1 3=7x1 4=9x1");
                 SensorParameterChange(LsPar, &LsPar->SpeckleFilter,      1, "Speckle Filter: 0=no filter 1=3x3 2=5x5 3=7x7 4=9x9");
                 SensorParameterChange(LsPar, &LsPar->SkipNoLaserPoints,  1, "Skip no laser detections points");
                 SensorParameterChange(LsPar, &LsPar->LaserMaskFilter,    1, "Laser Mask Filter");
                 // SensorParameterChange(LsPar, &LsPar->CommandModeASCII,   1, "Command transfer mode (0=Binary / 1=ASCII)");
                 break;
                 
      case 'r' :
                 for(PN=0; PN<2; PN++) 
                 {
                   TcpRecveCnt[PN] = 0; 
                   TcpSendCnt [PN] = 0; 
                   SendCounter[PN] = 0;
                   NbrLines   [PN] = 0;
                 }
                 break;

      case 'd' :
                 ChangeValue_Telnet(&LsPar->DebugPrint,       0,  9,  1, "Debug Print: 1=Memory 2=fps 3=Shutter 4=ImgAdr 5=TCP/IP 6=ImgToWorld 7=ProdRes 8=ParSet 9=SendData");
                 ChangeValue_Telnet(&LsPar->DebugDraw,        0,  3,  1, "Debug Draw: 1=LaserCenter 2=LaserDetection 3=ProductResults");
                 ChangeValue_Telnet(&LsPar->PrintHostCommand, 0, 15,  1, "Debug Commands: Bit0=Host->Cam Bit1=Cam->Host Bit2=SendData Bit3=GlobalHeader");
                 break;
                 
      case 'l' :
                 print("Press 'y' to load Sensor Parameter:\n");
                 if(ReadKeyboard_Telnet(1) == 'y') SensorParameterExchange(LsPar, PARAMETER_LOAD);
                 
                 #ifdef PRODUCT_VERSION
                 print("Press 'y' to load Product Parameter:\n");
                 if(ReadKeyboard_Telnet(1) == 'y') ProductParameterExchange(LsPar, ProdPar, PARAMETER_LOAD);
                 #endif
                 break;
                 
      case 's' :
                 print("Press 'y' to save Sensor Parameter:\n");
                 if(ReadKeyboard_Telnet(1) == 'y') SensorParameterExchange(LsPar, PARAMETER_SAVE);

                 #ifdef PRODUCT_VERSION
                 print("Press 'y' to save Product Parameter:\n");
                 if(ReadKeyboard_Telnet(1) == 'y') ProductParameterExchange(LsPar, ProdPar, PARAMETER_SAVE);
                 #endif
                 break;
                 
      case 'w' :
                 SensorParameterExchange(LsPar, PARAMETER_SHOW);
                 #ifdef PRODUCT_VERSION
                 ProductParameterExchange(LsPar, ProdPar, PARAMETER_SHOW);
                 #endif
                 break;
                 
                 #ifdef PRODUCT_VERSION
      case 'u' :
                 {
                 // 0 = Matrize  1 = Blech
                 static I32 JobNr = 1;
                 I32    temp;
        
                 // test with 
                 ChangeValue_Telnet(&OffsLaserWidth,     0,   64,  1, "OffsLaserWidth");;
                 ChangeValue_Telnet(&JobNr,                                  0,     1,  1, "Select Job Nr");
                 ChangeValue_Telnet(&ProdPar->LineAlgoMode[JobNr],           0,     2,  1, "LineAlgoMode: 0=BestLine 1=Hough 2=Combination (only at LineDetectionMode 0, 5 and 6)");
                 ChangeValue_Telnet(&ProdPar->LineDetectionMode[JobNr],      0,     6,  1, "LineDetectionMode: 0=AllPoints 1=LeftROI 2=RightROI 3=BestAngRotP 4=BestDisRotP 5=PreSelectMinDisRotP 6=PreSelectAngleRotP");
                 ChangeValue_Telnet(&ProdPar->HistoLineFilter  [JobNr],      0,    16,  1, "HistoLineFilter [Grad]");
                 temp = Round(10 * ProdPar->MaxLineDisSelect [JobNr]); ChangeValue_Telnet(&temp, 0,   500,  1, "MaxLineDisSelect  (1/10 mm)"); ProdPar->MaxLineDisSelect [JobNr] = (float)temp/10.0;
                 temp = Round(10 * ProdPar->MaxLineDisBestFit[JobNr]); ChangeValue_Telnet(&temp, 0,   500,  1, "MaxLineDisBestFit (1/10 mm)"); ProdPar->MaxLineDisBestFit[JobNr] = (float)temp/10.0;
                 temp = Round(ProdPar->RotationPointX_MM[JobNr]); ChangeValue_Telnet(&temp,  -1000,  1000,  1, "RotationPointX_MM"); ProdPar->RotationPointX_MM[JobNr] = temp;
                 temp = Round(ProdPar->RotationPointZ_MM[JobNr]); ChangeValue_Telnet(&temp,  -1000,  1000,  1, "RotationPointZ_MM"); ProdPar->RotationPointZ_MM[JobNr] = temp;
                 ChangeValue_Telnet(&ProdPar->MaxHoughLineNr   [JobNr],      1,     8,  1, "MaxHoughLineNr");
                 ChangeValue_Telnet(&ProdPar->MinHoughPointDis [JobNr],      1,    32,  1, "MinHoughPointDis");
                 ChangeValue_Telnet(&ProdPar->MaxHoughPointDis [JobNr],      1,   256,  1, "MaxHoughPointDis");
                 ChangeValue_Telnet(&DisplHoughLine,                        -2, ProdPar->MaxHoughLineNr[JobNr]-1,  1, "Display HoughLineNr: -2=LineDetec -1=NoDisp Nr:LineNr");
                 }
                 break; 
                 #endif
    }
  }
  while(response != 'q');


  /////////////////////////////////////////////
  // reset sensor with new settings
  /////////////////////////////////////////////

  SensorReset(LsPar);

  print("RUN mode\n");
}

/*************************************************************************************************************************/

void ChangeValue_Telnet (I32 *ParValue, I32 Min, I32 Max, I32 Step, I8 *text)
{
  I32  value;
  I32  response=0; 

  // start value
  value = *ParValue;

  // change value
  do
  {
    if(response == KEY_RIGHT || response == '+') {value+=Step; if(value > Max) value=Max;}
    if(response == KEY_LEFT  || response == '-') {value-=Step; if(value < Min) value=Min;}

    print("%s = %d  [%d - %d]   \r", text, value, Min, Max);

    response=ReadKeyboard_Telnet(1);
  }
  while( (response != KEY_ESC) && (response != KEY_ENTER) );

  // take new value
  if (response == KEY_ENTER)  *ParValue = value;

  print("%s = %d  [%d - %d]   \n", text, value, Min, Max);
}

/*************************************************************************************************************************/

#define TimeOut 10 /* ms */

int ReadKeyboard_Telnet (int wait)
{
  int i, ch1, ch2, ch3, ch4, src;


  if(wait || rbready())
  {

  while(1)
  {
    // if (kbready()) {src=1; break;}
    if (rbready()) {src=0; break;}
    time_delay(1);
  }

  if(src) ch1 = kbdrcv();
  else    ch1 = rs232rcv();

  if (ch1 != 0x1b)
     if(ch1 == 0x0d || ch1 == 0x0a)
       return(KEY_ENTER);
     else
       return(ch1);

  ch1 = KEY_ESC;

  for (i=0; i<TimeOut; i++)
  {
    time_delay(1);
    if(src) {if (kbready()) break;}
    else    {if (rbready()) break;}
  }
  if (i==TimeOut) return(ch1);  /* we've been waiting for too long */

  if(src) ch2 = kbdrcv();
  else    ch2 = rs232rcv();

  for (i=0; i<TimeOut; i++)
  {
    time_delay(1);
    if(src) {if (kbready()) break;}
    else    {if (rbready()) break;}
  }
  if (i==TimeOut) return(ch1);  /* we've been waiting for too long */

  if(src) ch3 = kbdrcv();
  else    ch3 = rs232rcv();


  if (ch2 == 0x4f)   /* 2/3 of a F1 or F2 received? */
  {

    if ((ch3 == 0x71) || (ch3 == 0x50))   /* F1 received? */
       return(KEY_F1);

    if ((ch3 == 0x72) || (ch3 == 0x51))   /* F2 received? */
       return(KEY_F2);

    return(KEY_INVALID);
  }

  if (ch2 == 0x5B)   /* 2/3 of a Cursor received? */
  {
    if (ch3 == 0x31)
    {
      for (i=0; i<TimeOut; i++)
      {
        time_delay(1);
        if(src) {if (kbready()) break;}
        else    {if (rbready()) break;}
      }
      if (i==TimeOut) return(ch1);  /* we've been waiting for too long */

      if(src) ch4 = kbdrcv();
      else    ch4 = rs232rcv();

      if (ch4 == 0x31)   /* F1 received? */
       return(KEY_F1);

      if (ch4 == 0x32)   /* F2 received? */
       return(KEY_F2);
    }
    else
    {
      switch (ch3)
    {
      case 0x41 :
      case 0x42 :
      case 0x43 :
      case 0x44 : return(ch3 - 61);
      default   : return(KEY_INVALID);
    }
     }
  }
  return(KEY_INVALID);
  }
  else
  {
  return(KEY_INVALID);
  }
}

/************************************************************************************************************************/

I32 ReadyForCalc (LsParameter *LsPar)
{
  if(LsPar->ExposureMode == DOUBLE_EXPOSURE)
  {
    if (LsPar->DoubleShutterImg)
    {
      LsPar->DoubleShutterImg = 0;

      return 1;
    }
    else
    {
      LsPar->DoubleShutterImg = 1;

      return 0;
    }
  }
  else
  {
    return 1;
  }
}

/***************************************************************************************************************************/

I32 ScreenBuffer2 (LsParameter *LsPar, enum ScrMode mode)
{
  I32        NrOfScreens = 2, Trigger = 0;
  static I32 Screen[2], TrackNr[2], ScrAdr=0, TakeFirstImage = 0, TakeNextImage = 0, calc, capt;

  // print("ScreenBuffer2:  Mode=%d\n", mode);

  if(LsPar->Trigger > 0) Trigger = 1;

  // screen allocation mode
  if (mode == SCREEN_BUFF_INI)
  {
    Screen [0] = ScrGetCaptPage;

    ScrAdr     = (I32)DRAMScreenMalloc();
    if(ScrAdr == 0) return ERR_MEMORY;
    Screen [1] = (ScrAdr & 0xFFFFFC00) + 1024;

    // now reset
    mode = SCREEN_BUFF_RESET;
  }

  // reset sensor (cancel images, new format etc.)
  if (mode == SCREEN_BUFF_RESET)
  {
    TrackNr[0] = getvar(IMGREADY) + 1;
    TrackNr[1] = getvar(IMGREADY) + 1;

    while(cancel_capture_rq());
    time_delay(50);
    SetImageFormat(LsPar);
    time_delay(50);

    // after SetImageFormat, the sensor needs a dummy image, as the first image still has the old format
    tpict();

    capt = 0;
    calc = 1;

    // print("RESET: TrackNr[%d]=%d\n", calc, TrackNr[calc]);

    // now ini variables
    mode = SCREEN_BUFF_FIRST;
  }


  // prepare sensor for first image
  if (mode == SCREEN_BUFF_FIRST)
  {
    TakeFirstImage = 1;
  }

  // change screen for calculation and acquisition
  if (mode == SCREEN_BUFF_SWAP)
  {
    I32 DeltaImg, Sh;

    // print("SWAP:  TrackNr[%d]=%d  IMGREADY=%d  Screen[%d]=%d\n", calc, TrackNr[calc], getvar(IMGREADY), capt, Screen[capt]);

    // take start image (fill image pipe)
    if(TakeFirstImage)
    {
      TakeFirstImage = 0;

      if( LsPar->ExposureMode == DOUBLE_EXPOSURE) Sh = LsPar->DoubleShutter[capt];
      else                                        Sh = LsPar->Shutter;

      LsPar->ShutterReal = shutter(Sh);

      TrackNr[capt] = capture_request(getvar(EXPCNT), getvar(GAIN), (I32 *)Screen[capt], Trigger);

      if(LsPar->DebugPrint == 4) print("Capt. Img 1: t=%3dms  CapPage=0x%08X  Sh=%4d  Trig=%d  Track[%d]=%3d  DoubleShutterImg=%d  NBR_LINES=%d\n", getvar(MSEC), Screen[capt], Sh, Trigger, capt, TrackNr[capt], LsPar->DoubleShutterImg, LsPar->NbrLines);

      // in auto trigger mode (no double exposure), put only one image in the image pipe
      if( (LsPar->Trigger < 0) && (LsPar->ExposureMode != DOUBLE_EXPOSURE) )
      {
        TakeNextImage  = -1; // next image at next call
      }
      else
      {
        TakeNextImage  =  1; // next image immediately
      }

      calc = capt;
      capt++;
      if (capt == NrOfScreens) capt=0;
    }


    // if it is the last image, don't put another one in the image pipe
    if(LsPar->NbrLines == 1)
    {
      // check, if it is the last image for both ports
      if( (NbrLines[0]==1 && NbrLines[1]<=1 && NbrLines[1]!=-1) || (NbrLines[1]==1 && NbrLines[0]<=1 && NbrLines[0]!=-1) )
      {
        if(LsPar->ExposureMode != DOUBLE_EXPOSURE)
        {
          TakeNextImage  = 0;
        }
        else
        {
          if(LsPar->DoubleShutterImg == 1) TakeNextImage  = 0;
        }
      }
    }


    // ready to take next image?
    if(TakeNextImage > 0)
    {
      TakeNextImage = 0;

      // in double exposure mode, take the second image immediately (don't wait for a trigger signal)
      if((LsPar->ExposureMode == DOUBLE_EXPOSURE) && (LsPar->DoubleShutterImg == 1) ) Trigger = 0;

      if( LsPar->ExposureMode == DOUBLE_EXPOSURE) Sh = LsPar->DoubleShutter[capt];
      else                                        Sh = LsPar->Shutter;

      LsPar->ShutterReal = shutter(Sh);

      TrackNr[capt] = capture_request(getvar(EXPCNT), getvar(GAIN), (I32 *)Screen[capt], Trigger);

      if(LsPar->DebugPrint == 4) print("Capt. Img X: t=%3dms  CapPage=0x%08X  Sh=%4d  Trig=%d  Track[%d]=%3d  DoubleShutterImg=%d  NBR_LINES=%d\n", getvar(MSEC), Screen[capt], Sh, Trigger, capt, TrackNr[capt], LsPar->DoubleShutterImg, LsPar->NbrLines);

      // queue is full (this error shoud not happen)
      if(TrackNr[capt] == 0) return SCREEN_RES_ERR_PIPE;
    }


    // important for mode auto trigger (no double exposure)
    if(TakeNextImage < 0) TakeNextImage = 1;


    // image ready for calculation?
    DeltaImg = TrackNr[calc] - getvar(IMGREADY);
    if( (DeltaImg <= 0) || (DeltaImg > NrOfScreens) )
    {
      TakeNextImage = 1;

      ScrSetLogPage(Screen[calc]);

      if(LsPar->DebugPrint == 4) print("Ready Image: t=%3dms  LogPage=0x%08X\n", getvar(MSEC), ScrGetLogPage);
      
      ////////////////////////////////////////////////////////////////////////////////////////////
      // generate TimeStamp (exposure ready) for every captured image
      ////////////////////////////////////////////////////////////////////////////////////////////
      {
        I32 i;
        imgts *ts_table = (imgts *)getvar(TIMESTAMP);
        
        for(i=0; i<IMGTS_SIZE && ts_table; i++, ts_table++)
      	{
          I32 imageno;
          // U64 exptimestamp;
          
          // disable interrupts
          // disints();
          
          imageno = ts_table->imageno;
          
          if(imageno == TrackNr[calc])
          {
            // exptimestamp = ts_table->exptimestamp;
            LsPar->ImgTimeStampMs = ts_table->exptimestamp;
            
            // print("%3d: ImgNr=%6d CalcImgNr=%6d ExpTime=%6d\n", i, imageno, TrackNr[calc], (U32)ts_table->exptimestamp);
            
            break;
          }
          
          // enable interrupts
          // enaints();
        }
      }
      

      calc = capt;
      capt++;
      if (capt == NrOfScreens) capt=0;

      return SCREEN_RES_IMAGE;
    }

    return SCREEN_RES_BUSY;
  }

  // release screen allocation mode
  if (mode == SCREEN_BUFF_DEINI)
  {
    if(Screen[0] != 0) ScrSetCaptPage(Screen[0]);

    while(cancel_capture_rq());

    // switch back to full image size
    LsPar->ROI_X  =  0;   
    LsPar->ROI_Y  =  0;   
    LsPar->ROI_DX =  ScrGetColumns;   
    LsPar->ROI_DY =  ScrGetRows;   

    time_delay(50);
    SetImageFormat(LsPar);
    time_delay(50);

    tpict();

    if(ScrAdr) DRAMPgFree(ScrAdr);
  }

  return SCREEN_RES_END;
}

/*************************************************************************************************************************/

I32 AutoShutter (LsParameter *LsPar)
{
  I32 i, Cnt=0, Nr, ShOffsPulseUs = 0;
  U32 AvgVal=0, NewSh, AutoShLast;
  U8  *restrict pLaserVal;
  static  I32 AutoShLast2, AutoShLast1 = -1;
  
  // at PULSE mode, we have a shutter offset of 20 s (other setup to get a null point line)
  if(LsPar->Laser == 2) ShOffsPulseUs = 20;

  // presetting
  if(AutoShLast1 < 0)  {AutoShLast2 = AutoShLast1 = LsPar->ShutterReal;}

  // calc new shutter time
  
  // if we have Product ROIs use only the Laser Values from the ROIs
  if(LsPar->NrLaserVal_ROIs > 8) {Cnt = LsPar->NrLaserVal_ROIs; pLaserVal = LsPar->pLaserVal_ROIs; LsPar->NrLaserVal_ROIs=0;}
  else                           {Cnt = LsPar->LaserPoints;     pLaserVal = LsPar->pLaserVal;}
  
  // for(i=0; i<Nr; i++) if(LsPar->pLaserVal[i]) print("%3d: %3d\n", i, (I32)LsPar->pLaserVal[i]);

  if(Cnt > 8)
  {
    // use alle laser values
    if(LsPar->AutoShutterFilter == 0)
    {
      AvgVal = FL_sum_U8P(Cnt, pLaserVal) / Cnt;
    }
    
    // use 50% center laser values
    if(LsPar->AutoShutterFilter == 4)
    {
      Cnt /= 2;
      AvgVal = FL_sum_U8P(Cnt, pLaserVal + Cnt/2) / Cnt;
    }
    
    // use 33% left laser values
    if(LsPar->AutoShutterFilter == 5)
    {
      Cnt /= 3;
      AvgVal = FL_sum_U8P(Cnt, pLaserVal) / Cnt;
    }
    
    // use 33% right laser values
    if(LsPar->AutoShutterFilter == 6)
    {
      Nr = Cnt;
      Cnt /= 3;
      AvgVal = FL_sum_U8P(Cnt, pLaserVal + Nr - Cnt) / Cnt;
    }
 
    // with histogramm 
    if(LsPar->AutoShutterFilter >= 1 && LsPar->AutoShutterFilter <= 3)
    {
      I32   SumAvg=0, SumNr=0;
      U16   Histo0[256], Histo1[256], Histo2[256], Histo3[256];
      
      FL_Set_U16P_U16(256, Histo0, 0);    
      FL_Set_U16P_U16(256, Histo1, 0);    
      FL_Set_U16P_U16(256, Histo2, 0);    
      FL_Set_U16P_U16(256, Histo3, 0);    

      FL_histo_U8P(Cnt, pLaserVal, Histo0, Histo1, Histo2, Histo3);      
      
      // use 50% median laser values
      if(LsPar->AutoShutterFilter == 1)
      {
        for(i=1; i<255; i++)
        {
          Nr = Histo0[i] + Histo1[i] + Histo2[i] + Histo3[i];
          
          SumNr  += Nr;
          
          if(SumNr > Cnt/4) break;
        }
        
        SumNr = 0;
        
        for(; i<255; i++)
        {
          Nr = Histo0[i] + Histo1[i] + Histo2[i] + Histo3[i];
          
          SumAvg += Nr * i;
          SumNr  += Nr;
          
          if(SumNr > Cnt/2) break;
        }
      }

      // use 50% bright laser values
      if(LsPar->AutoShutterFilter == 2)
      {
        for(i=255; i>=1; i--)
        {
          Nr = Histo0[i] + Histo1[i] + Histo2[i] + Histo3[i];
          
          SumAvg += Nr * i;
          SumNr  += Nr;
          
          if(SumNr > Cnt/2) break;
        }
      }
      
      // use 50% dark laser values
      if(LsPar->AutoShutterFilter == 3)
      {
        for(i=1; i<255; i++)
        {
          Nr = Histo0[i] + Histo1[i] + Histo2[i] + Histo3[i];
          
          SumAvg += Nr * i;
          SumNr  += Nr;
          
          if(SumNr > Cnt/2) break;
        }
      }

      AvgVal = SumAvg / SumNr;
    }
  }
  else
  {
    I32   MinLaserArea, LaserArea;
    image ImgArea;
    
    ImageAssign(&ImgArea, ScrByteAddr(0, 0), LsPar->ROI_DX, LsPar->ROI_DY_SubSample, ScrGetPitch);

    // No laser detected. Use image brightness in order to define, if image is to bright or to dark for laser detection
    // AvgVal = mean(&ImgArea); if(AvgVal > LsPar->RlcThresh) AvgVal = 255; else AvgVal=0;
    
    // new solution: If no laser points because of a to bright and wide line, the image brightness gives wrong results
    
    MinLaserArea = LsPar->ROI_DX * LsPar->MinWidth;
    LaserArea    = arx(&ImgArea, LsPar->RlcThresh-1);
    
    // faster, but no rlc available in main yet
    // LaserArea    =  rlc_area(U16 *rlc, -1); 
    
    if(LaserArea > MinLaserArea) AvgVal = LsPar->AutoShutterVal * 2; // next time half shutter 
    else                         AvgVal = LsPar->AutoShutterVal / 2; // next time double shutter 
  }

  AutoShLast = LsPar->ShutterReal;

  // The 6210 model doesn't change the shutter immediately. It takes one additional image.
  if(getvar(MODEL) == 6210) AutoShLast = AutoShLast2;
  
  // at PULSE mode, we have a shutter offset of 20 s (other setup to get a null point line)
  AutoShLast -= ShOffsPulseUs; if(AutoShLast < 10) AutoShLast = 10;

  if(AvgVal) NewSh = (LsPar->AutoShutterVal * AutoShLast) / AvgVal;
  else       NewSh =  LsPar->AutoShutterMax;
  
  // at PULSE mode, we have a shutter offset of 20 s (other setup to get a null point line)
  NewSh += ShOffsPulseUs;

  if(NewSh < LsPar->AutoShutterMin) NewSh = LsPar->AutoShutterMin;
  if(NewSh > LsPar->AutoShutterMax) NewSh = LsPar->AutoShutterMax;

  if(LsPar->DebugPrint == 3) print("Shutter: Last2=%5d Last1=%5d Last=%5d New=%5d us  /  Val=%3d (%d)  /  Cnt=%4d (%4d) Filter=%d\n", AutoShLast2, AutoShLast1, LsPar->ShutterReal, NewSh, AvgVal, LsPar->AutoShutterVal, Cnt, LsPar->LaserPoints, LsPar->AutoShutterFilter);

  AutoShLast2 = AutoShLast1;
  AutoShLast1 = LsPar->ShutterReal;

  LsPar->Shutter  = NewSh;

  return 0;
}

/************************************************************************************************************************/

I32 AutoTrigger (LsParameter *LsPar, enum TrigMode Mode)
{
  I32 TimeDiffUs;
  I64 TimeUs;
  static I64 AutoTrigTimeUs;


  // calc absolute time in s
  TimeUs = (I64)1000 * GetTimeMs() + TIMER0_DIG / TIMER0_US_DIV;


  if(Mode == AUTO_TRIG_INI)
  {
    AutoTrigTimeUs = TimeUs + (10000000 / LsPar->AutoTriggerFPS);

    if(LsPar->DebugPrint == 4) print("AutoTrig   INI: Time=%8dus  NextTrig=%8dus  Err=%3d  (NrLines=%3d)\n", (I32)(TimeUs % 100000000), (I32)(AutoTrigTimeUs % 100000000), LsPar->AutoTriggerError, LsPar->NbrLines);
  }


  if(Mode == AUTO_TRIG_DIFF)
  {
    TimeDiffUs = (I32)(TimeUs - AutoTrigTimeUs);

    // allow other tasks or interrupts
    if(TimeDiffUs < -1200) time_delay(1);

    // start trigger image
    if(TimeDiffUs >= -50)
    {
      // wait for the exact AutoTrigTimeUs
      while (TimeDiffUs < 0)
      {
        TimeUs = (I64)1000 * GetTimeMs() + TIMER0_DIG / TIMER0_US_DIV;
        TimeDiffUs = (I32)(TimeUs - AutoTrigTimeUs);
      }

      AutoTrigTimeUs += (10000000 / LsPar->AutoTriggerFPS);

      if (TimeDiffUs > MAX_JITTER_US) LsPar->AutoTriggerError++;

      if(LsPar->DebugPrint == 4) print("AutoTrig READY: Time=%8dus  NextTrig=%8dus  Err=%3d  (NrLines=%3d / TimeDiff=%3d)\n", (I32)(TimeUs % 100000000), (I32)(AutoTrigTimeUs % 100000000), LsPar->AutoTriggerError, LsPar->NbrLines, TimeDiffUs);

      return AUTO_TRIG_READY;
    }
  }

  return 0;
}

/*************************************************************************************************************************/

I32 StoreImage (void)
{
  static I32 ImgCount=0;
  I32    Error;
  char   ImgName[80];
  image  Source;

  ImageAssign(&Source, (U8 *)ScrGetLogPage, ScrGetColumns, ScrGetRows, ScrGetPitch);
  
  { 
    ChangeValue_Telnet(&ImgCount, 0, 999, 1, "Write image number");
    
    // #define JPG_IMG
    
    #ifdef JPG_IMG
      ////////////////////
      // write jpg image
      ////////////////////

      sprint(ImgName, "fd:/LS%03d.jpg",    ImgCount++);
      print("Write image '%s'. Please wait...\n", ImgName);
      Error = fwrite_jpeg(&Source, ImgName, 50, (I32)mremain());
    #else
      ////////////////////
      // write bmp image
      ////////////////////

      sprint(ImgName, "fd:/LS%03d.bmp",    ImgCount++);
      print("Write image '%s'. Please wait...\n", ImgName);
      Error = fwrite_image(ImgName, &Source); 
    #endif

    if (Error) print("Write image Error %d.\n", Error);
    else       print("Write image ready.\n");
  }
  
  return Error;
}

/*************************************************************************************************************************/

I32 ReadImage (void)
{
  static I32 ImgCount=0;
  I32  Error;
  char ImgName[80];
  image Source;

  // take image
  tpict();

  // delete old image
  ImageAssign(&Source, (U8 *)ScrGetLogPage, ScrGetColumns, ScrGetRows, ScrGetPitch);
  set(&Source, 0);

  // if (DmPar->DmError != 0)
  { 
    ChangeValue(&ImgCount, 0, 999, 1, "Read image number");

    sprint(ImgName, "fd:/LS%03d.bmp", ImgCount);
    print("Read image '%s'. Please wait...\n", ImgName);

    ////////////////////
    // read bmp image
    ////////////////////

    Error = fread_image(ImgName, &Source); 
     
    if (Error == 0) 
    {
      print("Read image ready.\n");
    }
    else
    {
      if (Error == ERR_OPEN) print("File '%s' not found.\n", ImgName);
      else                   print("Read BMP Typ Error %d.\n", Error);
    }
  }

  return Error;
}

/***************************************************************************************************************************/

void SensorReset (LsParameter *LsPar)
{
  static I32 GainVal   = -1000;
  static I32 Laser     = -1000;
  static I32 Trigger   = -1000;
  static I32 SubSample = -1000;  
  static I32 ROI_X     = -1000;
  static I32 ROI_Y     = -1000;
  static I32 ROI_DX    = -1000;
  static I32 ROI_DY    = -1000;  
  

  // reset double exposer image
  LsPar->DoubleShutterImg = 0;


  // set gain
  if(GainVal != LsPar->GainVal)  
  {
    GainVal = LsPar->GainVal;
    setvar(GAIN, LsPar->GainVal);
  }


  // set laser
  if(Laser != LsPar->Laser)  
  {
    Laser = LsPar->Laser;
    
    if((getvar(MODEL) / 100) == 62) SwitchLaser(LsPar->Laser);
  }


  // set trigger mode
  if(Trigger != LsPar->Trigger)  
  {
    Trigger = LsPar->Trigger;
    
    if(LsPar->Trigger == TRIGGER_RISING_EDGE ) TRIGINP_POS();
    if(LsPar->Trigger == TRIGGER_FALLING_EDGE) TRIGINP_NEG();

    // reset image format
    ScreenBuffer2(LsPar, SCREEN_BUFF_RESET);
  }


  // RESET_SENSOR (every cancel_capture_rq() will reset the image format!)
  if(SubSample != LsPar->SubSample || ROI_X != LsPar->ROI_X || ROI_Y != LsPar->ROI_Y || ROI_DX != LsPar->ROI_DX || ROI_DY != LsPar->ROI_DY)  
  {
    SubSample = LsPar->SubSample;
    
    ROI_X     = LsPar->ROI_X;
    ROI_Y     = LsPar->ROI_Y;
    ROI_DX    = LsPar->ROI_DX;
    ROI_DY    = LsPar->ROI_DY;
    
    // reset image format
    ScreenBuffer2(LsPar, SCREEN_BUFF_RESET);

    // set ROI range in mm
    // printf("SensorReset:\n");
    ImgToWorld_ROI(LsPar);
  }
}

/************************************************************************************************************************/

#define C6210_LCON              0x44000000          /* serial shift register for LCON     */

I32 SwitchLaser (I32 LaserOn)
{
  if(LaserOn)
  {
    // printf("switch on laser\n");

    // set current
    *((volatile char *)C6210_TTL)  = 0;     /* write access to different register clears aflg */
    *((volatile char *)C6210_LCON) = 0xFF;  /* data register    */
    *((volatile char *)C6210_LCON) = 0x00;  /* address register */

    time_delay(2);

    // set channel
    *((volatile char *)C6210_TTL)  = 0;     /* write access to different register clears aflg */
    if(LaserOn == 2) *((volatile char *)C6210_LCON) = 0x0F;  /* control register - pulse */
    else             *((volatile char *)C6210_LCON) = 0x1F;  /* control register - none pulse  */
    *((volatile char *)C6210_LCON) = 0x01;  /* address register   */
  }
  else
  {
    // printf("switch off laser\n");

    // set current
    *((volatile char *)C6210_TTL)  = 0;     /* write access to different register clears aflg */
    *((volatile char *)C6210_LCON) = 0x00;  /* data register    */
    *((volatile char *)C6210_LCON) = 0x00;  /* address register */

    time_delay(2);

    // set channel
    *((volatile char *)C6210_TTL)  = 0;     /* write access to different register clears aflg */
    *((volatile char *)C6210_LCON) = 0x00;  /* control register   */
    *((volatile char *)C6210_LCON) = 0x01;  /* address register   */
  }

  return 0;
}

/***************************************************************************************************************************/

I32 SetImageFormat (LsParameter *LsPar)
{
  I32 SubsamplingFactor=1, mode=0; // 0=bw 1=color sensor
  I40 NewShutter;

  LsPar->ROI_Y_SubSample  = LsPar->ROI_Y; 
  LsPar->ROI_DY_SubSample = LsPar->ROI_DY; 
  
  if( (getvar(MODEL) / 100) == 62)
  {
    // work around for correct image acquisition
    LsPar->ROI_DX = (LsPar->ROI_DX / 4) * 4;

    // print("Set Image Format (SubSample=%d) X=%3d Y=%3d DX=%d DY=%3d SH=%4d -> ", LsPar->SubSample, LsPar->ROI_X, LsPar->ROI_Y, LsPar->ROI_DX, LsPar->ROI_DY, LsPar->Shutter);

    // set image format (every cancel_capture_rq() will reset the format!)
    if(getvar(MODEL) == 6211)
    {
      if(LsPar->SubSample) SubsamplingFactor = (1 << LsPar->SubSample);
    
      set_image_format_ev76c560_subs(SubsamplingFactor, LsPar->ROI_X, LsPar->ROI_Y, LsPar->ROI_DX, LsPar->ROI_DY, LsPar->Shutter, mode, &LsPar->ROI_X, &LsPar->ROI_Y, &LsPar->ROI_DX, &LsPar->ROI_DY, &NewShutter); 
      
      LsPar->ROI_Y_SubSample  = LsPar->ROI_Y  / SubsamplingFactor; 
      LsPar->ROI_DY_SubSample = LsPar->ROI_DY / SubsamplingFactor; 
    }
    else
    {
      vcrt_set_image_format(LsPar->ROI_X, LsPar->ROI_Y, LsPar->ROI_DX, LsPar->ROI_DY, LsPar->Shutter, mode, &LsPar->ROI_X, &LsPar->ROI_Y, &LsPar->ROI_DX, &LsPar->ROI_DY, &NewShutter);

      LsPar->SubSample        = 0;
      LsPar->ROI_Y_SubSample  = LsPar->ROI_Y; 
      LsPar->ROI_DY_SubSample = LsPar->ROI_DY; 
    }

    // print("X=%3d Y=%3d DX=%d DY=%3d SH=%4d\n", LsPar->ROI_X, LsPar->ROI_Y, LsPar->ROI_DX, LsPar->ROI_DY, (I32)NewShutter);
  }

  return 0;
}

/************************************************************************************************************************/

I32 SetCommandPar(LsParameter *LsPar, I32 *Parameter, I32 *MinVal, I32 *MaxVal, I32 *NewVal, I32 CmdType)
{
  I32 SetVal, Error = ERR_NONE;

  // get range
  SensorParameterRange(LsPar, Parameter, &SetVal, MinVal, MaxVal);

  // SET command
  if(CmdType == 2)
  {
    if (*NewVal < *MinVal) Error = ERR_PARM_RANGE_MIN;
    if (*NewVal > *MaxVal) Error = ERR_PARM_RANGE_MAX;
  
    if(Error == ERR_NONE) *Parameter = *NewVal;
  }

  *NewVal = *Parameter;

  return Error;
}

/************************************************************************************************************************/

/************************************************************************************************************************/
//  TCP / IP - FUNCTIONS
/************************************************************************************************************************/

/************************************************************************************************************************/

I32 TCP_ReadCommand (LsParameter *LsPar, U32 *Alias, U8 *pCommand, I32 MinCommand, I32 MaxCommand)
{
  I32        Error, DataLen, HeaderLen, HeaderOffs, CheckSumData, Len;
  I32        TimeOutMs = TCP_TIMEOUT, TimeOutMsTmp;

  static I32 ReadLen = 0, MaxWrite;
  static U8  *pWrite, *pRead;
  
  /////////////////////////////////////////////////////
  // Command = Header (HeaderLen) + Data (DataLen)
  /////////////////////////////////////////////////////
  
  
  // ini values
  HeaderLen = MinCommand;  // GLOBAL_HEAD_SIZE
  
  if(ReadLen <= 0)
  {
    ReadLen  = 0;
    pWrite   = pCommand;
    pRead    = pCommand;
    MaxWrite = MaxCommand;
  }  
  
  
  // command reading loop
  do
  {
    // read at least HeaderLen (= gloabal header size) 
    if(ReadLen < HeaderLen)
    {
      Error = TCP_ReadData(Alias, pWrite, (HeaderLen - ReadLen), MaxWrite, TimeOutMs, &Len);

      // adapt pointer and size
      ReadLen  += Len;
      pWrite   += Len;
      MaxWrite -= Len;  
      
      // check, if HeaderLen Bytes are available
      if(Error < 0) return Error;
    }
    
    // analyse global header
    Error = GetGlobalHeader(pRead, ReadLen, &Error, &TimeOutMsTmp, &CheckSumData, &DataLen, &HeaderOffs, LsPar->PrintHostCommand & 0x08);

    // adapt pointer and size of available data
    pRead   += HeaderOffs;
    ReadLen -= HeaderOffs;

    // check for valid global header (GLOBAL_HEAD_SYNC)
    if(Error < 0) return Error;
    
    // read missing data
    if(ReadLen < DataLen) 
    {
      // read missing data
      Error = TCP_ReadData(Alias, pWrite, (DataLen - ReadLen), MaxWrite, TimeOutMs, &Len);
      
      // adapt pointer and size
      ReadLen  += Len;
      pWrite   += Len;
      MaxWrite -= Len;  

      // check, if missing data are now available
      if(Error < 0) return Error;
    }    
    
    // perform command
    if(DataLen > 0) 
    {
      TCP_RunCommand(LsPar, pRead, DataLen, *Alias);

      // adapt pointer and size
      pRead   += DataLen;
      ReadLen -= DataLen;
    }

    // copy data to pCommand for full command space
    if(ReadLen > 0) FL_CPY_U8P_U8P(ReadLen, pRead, pCommand); 
    
    // set pointer and size
    pRead    = pCommand;
    pWrite   = pCommand   + ReadLen;
    MaxWrite = MaxCommand - ReadLen;
  }
  while(ReadLen > 0); // more commands?
  
  
  return 0;  
}

/************************************************************************************************************************/

I32 TCP_RunCommand (LsParameter *LsPar, U8 *pCmdData, I32 CmdLen, U32 Alias)
{
  I32   Error, CommandError = ERR_TCP_IP;
  I32   MinVal, MaxVal, NewVal;
  volatile U32   CommandType, ParameterID, HostCounter, StringLen;
  U8   *pCommand;


  /////////////////////////////////////////////////////////
  // ini values
  ///////////////////////////////////////////////////////// 

  Error    = 0;
  MinVal   = 0;
  MaxVal   = 0;
  NewVal   = 0;
  pCommand = pCmdData;


  /////////////////////////////////////////////////////////
  // Read Command Data
  ///////////////////////////////////////////////////////// 

  // <CommandType> + <ParameterID> + <HostCounter> for STRING, GET and SET commands
  FL_CPY_U8P_U8P(4, pCommand, (U8 *)&CommandType); pCommand += 4; 
  FL_CPY_U8P_U8P(4, pCommand, (U8 *)&ParameterID); pCommand += 4; 
  FL_CPY_U8P_U8P(4, pCommand, (U8 *)&HostCounter); pCommand += 4; 

  // additional data for STRING command
  if (CommandType == 0) 
  {
    // get StringLen
    FL_CPY_U8P_U8P(4, pCommand, (U8 *)&StringLen); pCommand += 4;
    
    // print("CommandType = 0: StringLen = %d\n", StringLen);
    
    if(StringLen > 0 && StringLen < MAX_SETTING_SIZE-1) // -1 to terminate string
    {
      #ifdef PRODUCT_VERSION
      if(ParameterID == CMD_RECEIVE_PRODUCT_DATA)
      {
        // get StringData
        FL_CPY_U8P_U8P(StringLen, pCommand, (U8 *)ProdPar->pProductSetting); 
        
        // terminate string
        ProdPar->pProductSetting[StringLen] = 0;
        ProdPar->ProdSettingSize = StringLen + 1;
      }
      #endif 
      
      if(ParameterID == CMD_RECEIVE_SENSOR_DATA)
      {
        // get StringData
        FL_CPY_U8P_U8P(StringLen, pCommand, (U8 *)LsPar->pSensorSetting); 
        
        // terminate string
        LsPar->pSensorSetting[StringLen] = 0;
        LsPar->SensorSettingSize = StringLen + 1;
      }
      
      pCommand += StringLen;
    }  
    else    
    {
      LsPar->pSensorSetting[0]    = 0;
      LsPar->SensorSettingSize    = 0;

      #ifdef PRODUCT_VERSION
      ProdPar->pProductSetting[0] = 0;
      ProdPar->ProdSettingSize    = 0;
      #endif

      if(StringLen != 0) Error = ERR_CMD_STRG_LEN;
    }  
  }

  // additional data for SET command
  if (CommandType == 2) 
  {
    FL_CPY_U8P_U8P(4, pCommand, (U8 *)&NewVal   ); pCommand += 4; 
  }


  /////////////////////////////////////////////////////////
  // send remaining data before acknowledge new command
  ///////////////////////////////////////////////////////// 

//  if(ParameterID == CMD_STOP) TCP_SendCommand(LsPar, Alias, LsPar->SendBuffer[PN], LsPar->MaxSendBuff, INI_SEND_BUFFER); // ???
  if(ParameterID == CMD_STOP) TCP_SendCommand(LsPar, Alias, LsPar->SendBuffer[PN], LsPar->MaxSendBuff, EMPTY_SEND_BUFFER);
  else                        TCP_SendCommand(LsPar, Alias, LsPar->SendBuffer[PN], LsPar->MaxSendBuff, EMPTY_SEND_BUFFER);


  ///////////////////////////////////////////////////////// 
  // check command
  ///////////////////////////////////////////////////////// 

  if(CommandType  >                   2) Error = ERR_CMD_TYPE;
  if((I32)(pCommand-pCmdData) != CmdLen) Error = ERR_CMD_DATA_LEN;
    
  // print host string
  if (LsPar->PrintHostCommand & 0x01)
  {
    print("Host->Cam Port %d (Host Cmd): ", PN); 
    print("TcpRecveCnt=%4d  CommandType=%d  ParameterID=%2d  HostCounter=%4d  ", TcpRecveCnt[PN]-1, CommandType, ParameterID, HostCounter);
    if(CommandType == 0) print("StrLen=%4d  ", StringLen);
    if(CommandType == 2) print("SetVal=%4d  ", NewVal);
    print("CmdLen=%4d  Error=%d\n", CmdLen, Error);
    if(CommandType == 0) if(ParameterID == CMD_RECEIVE_SENSOR_DATA ) print("SensStr=%s", (U8 *)LsPar->pSensorSetting);
    #ifdef PRODUCT_VERSION
    if(CommandType == 0) if(ParameterID == CMD_RECEIVE_PRODUCT_DATA) print("ProdStr=%s", (U8 *)ProdPar->pProductSetting);
    #endif
  }
  
  
  ///////////////////////////////////////////////////////// 
  // in error case, go direct to command response
  ///////////////////////////////////////////////////////// 

  if(Error) goto CMD_RESPONSE;
  
  

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // HOST -> VISION
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  switch(ParameterID)
  {
    case  CMD_STOP:
          TcpRecveCnt[PN] = 0; 
          TcpSendCnt [PN] = 0; 
          SendCounter[PN] = 0;
          NbrLines   [PN] = 0;
          CommandError = SetCommandPar(LsPar, &LsPar->NbrLines, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_LASER_MODE:
          CommandError = SetCommandPar(LsPar, &LsPar->Laser, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_ROI_X:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_X, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_ROI_Y:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_Y, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_ROI_DX:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_DX, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_ROI_DY:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_DY, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_GAIN:
          CommandError = SetCommandPar(LsPar, &LsPar->GainVal, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_SHUTTER:
          CommandError = SetCommandPar(LsPar, &LsPar->Shutter, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_TRIGGER:
          CommandError = SetCommandPar(LsPar, &LsPar->Trigger, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_MODE:
          CommandError = SetCommandPar(LsPar, &LsPar->DataMode, &MinVal, &MaxVal, &NewVal, CommandType);
          DataMode[PN] = LsPar->DataMode;
          break;

    case  CMD_NBR_LINES:
          CommandError = SetCommandPar(LsPar, &LsPar->NbrLines, &MinVal, &MaxVal, &NewVal, CommandType);
          NbrLines[PN] = LsPar->NbrLines;
          break;

    case  CMD_METHOD:
          CommandError = SetCommandPar(LsPar, &LsPar->Methode, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_ALGORITHM:
          CommandError = SetCommandPar(LsPar, &LsPar->Algorithm, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_THRESH:
          CommandError = SetCommandPar(LsPar, &LsPar->RlcThresh, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_MINWIDTH:
          CommandError = SetCommandPar(LsPar, &LsPar->MinWidth, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_MAXWIDTH:
          CommandError = SetCommandPar(LsPar, &LsPar->MaxWidth, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_EXPMODE:
          CommandError = SetCommandPar(LsPar, &LsPar->ExposureMode, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_LASERVAL:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoShutterVal, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_MINSH:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoShutterMin, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_MAXSH:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoShutterMax, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_DOUBLE_SH1:
          CommandError = SetCommandPar(LsPar, &LsPar->DoubleShutter[0], &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_DOUBLE_SH2:
          CommandError = SetCommandPar(LsPar, &LsPar->DoubleShutter[1], &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_FILTER:
          CommandError = SetCommandPar(LsPar, &LsPar->ReflexionFilter, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_ROI_X_MM:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_X_MM, &MinVal, &MaxVal, &NewVal, CommandType);
          if(CommandType == 1) WorldToImg_ROI(LsPar);
          break;

    case  CMD_ROI_Z_MM:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_Z_MM, &MinVal, &MaxVal, &NewVal, CommandType);
          if(CommandType == 1) WorldToImg_ROI(LsPar);
          break;

    case  CMD_ROI_DX_MM:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_DX_MM, &MinVal, &MaxVal, &NewVal, CommandType);
          if(CommandType == 1) WorldToImg_ROI(LsPar);
          break;

    case  CMD_ROI_DZ_MM:
          CommandError = SetCommandPar(LsPar, &LsPar->ROI_DZ_MM, &MinVal, &MaxVal, &NewVal, CommandType);
          if(CommandType == 1) WorldToImg_ROI(LsPar);
          break;

    case  CMD_LIB_VERSION:
          CommandError = SetCommandPar(LsPar, &LsPar->LibVersion, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_PLC_IN:
          LsPar->InPLC = NewVal = inPLC() & 0x0F;
          CommandError = SetCommandPar(LsPar, &LsPar->InPLC, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_PLC_OUT:
          CommandError = SetCommandPar(LsPar, &LsPar->OutPLC, &MinVal, &MaxVal, &NewVal, CommandType);
          if(CommandError == ERR_NONE) outPLC(NewVal);
          break;

    case  CMD_ETHERNET_PACKAGE:
          CommandError = SetCommandPar(LsPar, &LsPar->EthernetPackNr, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_AUTO_TRIG_PERIODE:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoTriggerFPS, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_AUTO_TRIG_ERR:
          CommandError = SetCommandPar(LsPar, &LsPar->AutoTriggerError, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_SEND_NOWAIT:
          CommandError = SetCommandPar(LsPar, &LsPar->EthernetSendNoWait, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_SUBSAMPLE:
          CommandError = SetCommandPar(LsPar, &LsPar->SubSample, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_CAMERA_SNR:
          CommandError = SetCommandPar(LsPar, &LsPar->CameraSNR, &MinVal, &MaxVal, &NewVal, CommandType);
          break;
          
    case  CMD_SPECKLE_FILTER:
          CommandError = SetCommandPar(LsPar, &LsPar->SpeckleFilter, &MinVal, &MaxVal, &NewVal, CommandType);
          break;

    case  CMD_MEDIAN_FILTER:
          CommandError = SetCommandPar(LsPar, &LsPar->MedianFilter, &MinVal, &MaxVal, &NewVal, CommandType);
          break;
          
    case  CMD_SAVE_SENSOR_FILE:
          CommandError = SensorParameterExchange(LsPar, PARAMETER_SAVE);
          NewVal = LsPar->SensorSettingSize;
          MinVal = 0;
          MaxVal = MAX_SETTING_SIZE;
          break;

    case  CMD_LOAD_SENSOR_FILE:
          CommandError = SensorParameterExchange(LsPar, PARAMETER_LOAD);
          NewVal = LsPar->SensorSettingSize;
          MinVal = 0;
          MaxVal = MAX_SETTING_SIZE;
          break;
          
    case  CMD_RECEIVE_SENSOR_DATA:
          CommandError = SensorParameterExchange(LsPar, PARAMETER_RECEIVE);
          break;
          
    case  CMD_SEND_SENSOR_DATA:
          CommandError = SensorParameterExchange(LsPar, PARAMETER_SEND);
          break;
          
    #ifdef PRODUCT_VERSION
    
    case  CMD_SAVE_PRODUCT_FILE:
          CommandError = ProductParameterExchange(LsPar, ProdPar, PARAMETER_SAVE);
          NewVal = ProdPar->ProdSettingSize;
          MinVal = 0;
          MaxVal = MAX_SETTING_SIZE;
          break;
          
    case  CMD_LOAD_PRODUCT_FILE:
          CommandError = ProductParameterExchange(LsPar, ProdPar, PARAMETER_LOAD);
          NewVal = ProdPar->ProdSettingSize;
          MinVal = 0;
          MaxVal = MAX_SETTING_SIZE;
          break;
          
    case  CMD_RECEIVE_PRODUCT_DATA:
          CommandError = ProductParameterExchange(LsPar, ProdPar, PARAMETER_RECEIVE);
          break;
          
    case  CMD_SEND_PRODUCT_DATA:
          CommandError = ProductParameterExchange(LsPar, ProdPar, PARAMETER_SEND);
          break;
          
    #endif          
          
    default:
          NewVal = 0;
          CommandError = ERR_COMMAND;
  }
  
  

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // create command response
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  CMD_RESPONSE:
  
  {
    I32 SendLen;

    // U8  Response[GLOBAL_HEAD_SIZE + 24 + MAX_SETTING_SIZE], *pResponse;
    U8  *Response, *pResponse;
    

    // ??? use send buffer for response, as we send already all data
    Response = LsPar->SendBuffer[PN]; 

    // write position after gloabal header
    pResponse = Response + GLOBAL_HEAD_SIZE;
    
    // use main error in response
    if(Error) CommandError = Error;
    
    // <CommandType> + <ParameterID> + <HostCounter> + <CommandError> for STRING, GET and SET response
    FL_CPY_U8P_U8P(4, (U8 *)&CommandType,  pResponse); pResponse += 4;
    FL_CPY_U8P_U8P(4, (U8 *)&ParameterID,  pResponse); pResponse += 4;
    FL_CPY_U8P_U8P(4, (U8 *)&HostCounter,  pResponse); pResponse += 4;
    FL_CPY_U8P_U8P(4, (U8 *)&CommandError, pResponse); pResponse += 4;

    // additional data for STRING response
    if(CommandType == 0)
    {
      U8  *pStringLen;
      
      // set output string length
      StringLen= 0;
      
      // U32 space for string length
      pStringLen = pResponse;
      pResponse += 4;
      
      // special command to send sensor parameter from Cam -> Host
      if(ParameterID == CMD_SEND_SENSOR_DATA)
      {
        StringLen = LsPar->SensorSettingSize;
        FL_CPY_U8P_U8P(LsPar->SensorSettingSize, (U8 *)LsPar->pSensorSetting, (U8 *)pResponse); pResponse += LsPar->SensorSettingSize;
      }

      // special command to send product parameter from Cam -> Host
      #ifdef PRODUCT_VERSION
      if(ParameterID == CMD_SEND_PRODUCT_DATA)
      {
        StringLen = ProdPar->ProdSettingSize;
        FL_CPY_U8P_U8P(ProdPar->ProdSettingSize, (U8 *)ProdPar->pProductSetting, (U8 *)pResponse);  pResponse += ProdPar->ProdSettingSize;
      }
      #endif

      // set string lenght
      FL_CPY_U8P_U8P(4, (U8 *)&StringLen, pStringLen);
    }

    // additional data for GET and SET response
    if(CommandType == 1 || CommandType == 2)
    {
      FL_CPY_U8P_U8P(4, (U8 *)&NewVal,       pResponse); pResponse += 4;
      FL_CPY_U8P_U8P(4, (U8 *)&MinVal,       pResponse); pResponse += 4;
      FL_CPY_U8P_U8P(4, (U8 *)&MaxVal,       pResponse); pResponse += 4;
    }

    

    ////////////////////////////////////
    // create Gloabal Header
    ////////////////////////////////////

    SendLen = (I32)(pResponse - Response);
    
    SetGlobalHeader(Response, SendLen - GLOBAL_HEAD_SIZE, 0, LsPar->PrintHostCommand & 0x08);


    ////////////////////////////////////
    // transfer response
    ////////////////////////////////////

    while(SendLen > 0)
    {
      I32 Len = TCP_Send(Alias, (char *)Response, SendLen);

      if(Len > 0) SendLen -= Len;
      else        SendLen  =  -1;

      // print("SendLen=%d Len=%d\n", SendLen, Len);
    }


    // print response string
    if (LsPar->PrintHostCommand & 0x02)
    {
      print("Cam->Host Port %d (Cmd Resp): ", PN);
      print("TcpSendCnt =%4d  CommandType=%d  ParameterID=%2d  HostCounter=%4d  Error=%d  ", TcpSendCnt[PN]-1, CommandType, ParameterID, HostCounter, CommandError);
      if(CommandType == 0                    ) print("StrLen=%4d  ", StringLen);
      if(CommandType == 1 || CommandType == 2) print("NewVal=%4d  MinVal=%4d  MaxVal=%4d", NewVal, MinVal, MaxVal);
      print("\n");
      if(CommandType == 0) if(ParameterID == CMD_SEND_SENSOR_DATA ) print("SensStr=%s", (U8 *)LsPar->pSensorSetting);
      #ifdef PRODUCT_VERSION
      if(CommandType == 0) if(ParameterID == CMD_SEND_PRODUCT_DATA) print("ProdStr=%s", (U8 *)ProdPar->pProductSetting);
      #endif
      print("\n");
    }
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // RESET_SENSOR - If necessary set sensor to new values (ROI, Shutter etc.)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  SensorReset(LsPar);

  return Error;
}

/************************************************************************************************************************/

I32 TCP_SendCommand (LsParameter *LsPar, U32 Alias, U8 *Command, I32 maxlen, enum SendMode Mode)
{
  I16    x, y, dx, dy;
  I32    i, count, SendSize, DataSize, MaxHeaderSize, BuffLen, Error=0;
  I32    SubHeader[8], SubHeaderLen;
  U8     *pGlobalHeaderStart;
  I32    ParVal_NbrLines, ParVal_DataMode;
  
  static U8  *pSend    [2];
  static U8  *pWrite   [2]; 
  static I32  PackCnt  [2];
  static I32  OptNoWait[2];

  // test 1460 Bytes Send Mode  
  // static U8 *pW, *pS;
  // print("Command=0x%x pSend=0x%x (0x%x) pWrite=0x%x (0x%x) Data=%6d Send=%6d Remain=%6d\n", Command, pSend[PN], pSend[PN] - (int)(pSend[PN] - pS), pWrite[PN], pWrite[PN] - (int)(pWrite[PN] - pW), (int)(pWrite[PN] - pW), (int)(pSend[PN] - pS), (int)(pWrite[PN] - pSend[PN]));
  // pW = pWrite[PN];
  // pS = pSend[PN];
  // print("SEND PN=%d Remain=%6d\n", PN, (int)(pWrite[PN] - pSend[PN]));
  // if(Mode != EMPTY_SEND_BUFFER) print("SEND PN=%d  Mode=%d\n", PN, Mode);
  

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // ini values
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  
  ParVal_NbrLines = NbrLines[PN];
  ParVal_DataMode = DataMode[PN];
  
  if(LsPar->DebugPrint == 9) print("FUNC: Alias=0x%08x PN=%d TcpSendCnt=%5d TcpRecveCnt=%4d Mode=%d RestData=%7d Mode=%d NrLines=%2d\n", Alias, PN, TcpSendCnt[PN], TcpRecveCnt[PN], Mode, (int)(pWrite[PN] - pSend[PN]), ParVal_DataMode, ParVal_NbrLines);

  if(Mode ==   INI_SEND_BUFFER) {pSend[PN] = pWrite[PN] = NULL; return 0;}
  if(Mode == EMPTY_SEND_BUFFER) {if( (I32)(pWrite[PN] - pSend[PN]) == 0) return 0;}
  
  if(pSend[PN]  ==              NULL) {pSend[PN] = pWrite[PN] = Command;}

  BuffLen = maxlen - (I32)(pWrite[PN] - Command);

  // check ethernet cable connection
  if(getvar(ETHLINK) == 0) return 0;
  
  // check, if Alias socket is available
  if(Alias == 0) return 0;


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // check parameter
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (ParVal_DataMode < 1) || (ParVal_DataMode > 9) ) return ERR_PARM;

  // print("ParVal_DataMode = %d\n", ParVal_DataMode);


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // calc size
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  DataSize      =  0;
  MaxHeaderSize = GLOBAL_HEAD_SIZE + 24; // GLOBAL_HEAD_SIZE + SubHeader in Bytes

  if(ParVal_DataMode ==  1) DataSize = MaxHeaderSize +     LsPar->ROI_DX * LsPar->ROI_DY_SubSample;
  if(ParVal_DataMode ==  2) DataSize = MaxHeaderSize +     LsPar->LaserPoints * sizeof(float);
  if(ParVal_DataMode ==  3) DataSize = MaxHeaderSize +     LsPar->LaserPoints * (sizeof(float) + 1);
  if(ParVal_DataMode ==  4) DataSize = MaxHeaderSize + 2 * LsPar->LaserPoints * sizeof(float);
  if(ParVal_DataMode ==  5) DataSize = MaxHeaderSize + 2 * LsPar->LaserPoints * sizeof(float);
  if(ParVal_DataMode ==  6) DataSize = MaxHeaderSize +     LsPar->ROI_DX * LsPar->ROI_DY_SubSample;
  if(ParVal_DataMode ==  7) DataSize = MaxHeaderSize +     LsPar->LaserPoints * (sizeof(float) + 1) + 256;
  if(ParVal_DataMode ==  8) DataSize = MaxHeaderSize +     MAX_PROD_RESULT_STRING;
  if(ParVal_DataMode ==  9) DataSize = MaxHeaderSize + 2 * LsPar->LaserPoints * sizeof(float) + MAX_PROD_RESULT_STRING;


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // optimize TPP/IP transfer (OPT_SEND_NOWAIT)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  
  {
    /* send data immediately (VCRT_MSG_NONBLOCK) */
    I32 opt_value = 1;

    if(LsPar->EthernetSendNoWait  == 0) opt_value = 0;
    if(ParVal_DataMode            == 1) opt_value = 0;
    if(ParVal_DataMode            == 6) opt_value = 0;
    if(BuffLen     <          DataSize) opt_value = 0;
    if(Mode       == EMPTY_SEND_BUFFER) opt_value = 0;

    if(opt_value != OptNoWait[PN])
    {
      OptNoWait[PN] = opt_value;

      Error = setsockopt(Alias, SOL_TCP, OPT_SEND_NOWAIT, &opt_value, sizeof(opt_value));

      // print("New opt_value = %d  Error=%d\n", opt_value, Error);
    }
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // if not enough memory for new data, send complete buffer
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (BuffLen < DataSize) || (Mode == EMPTY_SEND_BUFFER) )
  {
    // calc response len
    count = (I32)(pWrite[PN] - pSend[PN]);

    // send all data (opt_value must be 0 !)
    SendSize = TCP_Send(Alias, (char *)pSend[PN], count);

    if (LsPar->DebugPrint == 5) print("Buff SendSize=%6d Remain=%10d (Empty Buffer)\n", SendSize, count - SendSize);

    // print("Error TCP Send: Buffer Full\n");

    pSend[PN] = pWrite[PN] = Command;

    if(SendSize < 0) {if (LsPar->DebugPrint == 5) print("Error=%d TCP Send: Size=%d\n", SendSize, count); return -1;}
    if(Mode == EMPTY_SEND_BUFFER) {return 0;}
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // VISION -> HOST (write SendDate in Buffer)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // no TCP/IP transfer
  if(ParVal_DataMode <  0) return 0;
  if(ParVal_NbrLines == 0) return 0;


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Space for Gloabal Header and create Meta Header
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  {
    U32 CommandType;

    // remember Gloabal Header start
    pGlobalHeaderStart = pWrite[PN];
  
    // start position for data
    pWrite[PN] += GLOBAL_HEAD_SIZE;
  
    // DATA Result Response 
    CommandType = 100;

    // create META Header
    FL_CPY_U8P_U8P(4, (U8 *)&CommandType,           pWrite[PN]); pWrite[PN] += 4;
    FL_CPY_U8P_U8P(4, (U8 *)&ParVal_DataMode,       pWrite[PN]); pWrite[PN] += 4;
    FL_CPY_U8P_U8P(4, (U8 *)&SendCounter[PN],       pWrite[PN]); pWrite[PN] += 4;
    FL_CPY_U8P_U8P(8, (U8 *)&LsPar->ImgTimeStampMs, pWrite[PN]); pWrite[PN] += 8;

    SendCounter[PN]++;
  }  
  
  
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Create Data
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /*
  0. State MODE_STANDBY
  1. State MODE_TAKE_PIC
  2. State MODE_POINTLIST
  3. State MODE_POINTLIST_AND_LASER_IMG
  4. State MODE_POINTLIST_MM_XZ_XZ (x0 z0 / x1 z1 / x2 z2 / ...)
  5. State MODE_POINTLIST_MM_XX_ZZ (x0 x1 x2 ... / z0 z1 z2 ...)
  6. State MODE_BINARY_PIC
  7. State MODE_POINTLIST_AND_LASER_IMG + Product Angle Results
  8. State MODE_PRODUCT_RESULTS
  9. State MODE_POINTLIST_MM + Product Angle Results
  */

  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 1. State MODE_TAKE_PIC
  // 6. State MODE_BINARY_PIC
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (ParVal_DataMode == 1) || (ParVal_DataMode == 6) )
  {
    I32 NextLine;
    U8  *restrict pPix;

    // set "Response Sub Header"
    SubHeaderLen = 8;
    SubHeader[0] = LsPar->ROI_DX;
    SubHeader[1] = LsPar->ROI_DY_SubSample;
    FL_CPY_U8P_U8P(SubHeaderLen, (U8 *)SubHeader, pWrite[PN]); pWrite[PN] += SubHeaderLen;

    // copy image data
    pPix     = (U8 *)ScrGetLogPage;
    dx       =       LsPar->ROI_DX;
    dy       =       LsPar->ROI_DY_SubSample;
    NextLine =       ScrGetPitch - dx;

    for(y=0; y<dy; y++)
    {
      if(ParVal_DataMode == 1)
      {
        for(x=0; x<dx; x++)
        {
          *pWrite[PN]++ = *pPix++;
        }
      }
      else
      {
        for(x=0; x<dx; x++)
        {
          if(*pPix++ >= LsPar->RlcThresh) *pWrite[PN]++ = 255;
          else                            *pWrite[PN]++ =   0;
        }
      }

      pPix += NextLine;
    }
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 2. State MODE_POINTLIST
  // 3. State MODE_POINTLIST_AND_LASER_IMG
  // 7. State MODE_POINTLIST_AND_LASER_IMG + Product Angle Results
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (ParVal_DataMode == 2) || (ParVal_DataMode == 3) || (ParVal_DataMode == 7) )
  {
    I32 Nr=0;

    // set "Response Sub Header"
    SubHeaderLen = 16;
    SubHeader[0] = LsPar->Algorithm;
    SubHeader[1] = LsPar->ROI_DX / LsPar->StepDx;
    SubHeader[2] = LsPar->ROI_X;
    SubHeader[3] = LsPar->StepDx;
    FL_CPY_U8P_U8P(SubHeaderLen, (U8 *)SubHeader, pWrite[PN]); pWrite[PN] += SubHeaderLen;
    
    // set "Response DATA"
    if(LsPar->Algorithm)
    {
      float TempF32, Accuracy_Inv = 1.0 / (float)LASER_POS_MUL;
  
      for(i = LsPar->ROI_X; i < LsPar->ROI_X+LsPar->ROI_DX; i += LsPar->StepDx)
      {
        // in case of SkipNoLaserPoints, set missing laser points to NO_LASER_DETEC
        if ( (Nr < LsPar->LaserPoints) && ((i<<6) >= LsPar->pLaserPosX[Nr]) ) TempF32 = (float)LsPar->pLaserPosZ[Nr++] * Accuracy_Inv;
        else                                                                  TempF32 = (float)NO_LASER_DETEC          * Accuracy_Inv;
        
        FL_CPY_U8P_U8P(4, (U8 *)&TempF32, pWrite[PN]); pWrite[PN] += 4;
      }
    }
    else
    {
      I16  TempI16;
  
      for(i = LsPar->ROI_X; i < LsPar->ROI_X+LsPar->ROI_DX; i += LsPar->StepDx)
      {
        // in case of SkipNoLaserPoints, set missing laser points to NO_LASER_DETEC
        if ( (Nr < LsPar->LaserPoints) && ((i<<6) >= LsPar->pLaserPosX[Nr]) ) TempI16 = (I16)(LsPar->pLaserPosZ[Nr++] >> 2);
        else                                                                  TempI16 = (I16)(NO_LASER_DETEC          >> 2);
        
        FL_CPY_U8P_U8P(2, (U8 *)&TempI16, pWrite[PN]); pWrite[PN] += 2;
      }
    }

    if( (ParVal_DataMode == 3) || (ParVal_DataMode == 7) )
    {
       Nr = 0;
       
      // ADD Laser Brightness to "Response DATA"
      for(i = LsPar->ROI_X; i < LsPar->ROI_X+LsPar->ROI_DX; i += LsPar->StepDx)
      {
        // in case of SkipNoLaserPoints, set missing laser points to color 0
        if ( (Nr < LsPar->LaserPoints) && ((i<<6) >= LsPar->pLaserPosX[Nr]) ) *pWrite[PN]++ = LsPar->pLaserVal[Nr++];
        else                                                                  *pWrite[PN]++ = 0;
      }
    }
    

    #ifdef PRODUCT_VERSION
    if(ParVal_DataMode == 7)
    {
      // Product Angle Results
      {
        I32 Nr       = 2 * 4 * sizeof(float);
        U8  *pStruct = (U8 *)&ProdPar->ResultAngle[0];

        FL_CPY_U8P_U8P( 4, (U8 *)&Nr,      pWrite[PN]); pWrite[PN] += 4;
        FL_CPY_U8P_U8P(Nr, (U8 *)&pStruct, pWrite[PN]); pWrite[PN] += Nr;
      }
    }
    #endif
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 4. State MODE_POINTLIST_MM (x0 z0 / x1 z1 / x2 z2 / ...)
  // 5. State MODE_POINTLIST_MM (x0 x1 x2 ... / z0 z1 z2 ...)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  if( (ParVal_DataMode == 4) || (ParVal_DataMode == 5) )
  {
    I32   CountPoints = LsPar->LaserPoints;
    float CoordXZ[2];

    // set "Response Sub Header"
    SubHeaderLen = 4;
    SubHeader[0] = CountPoints;
    FL_CPY_U8P_U8P(SubHeaderLen, (U8 *)SubHeader, pWrite[PN]); pWrite[PN] += SubHeaderLen;
    
    // write world coordinates into transfer buffer
    if(ParVal_DataMode == 4)
    {
      for(i=0; i<CountPoints; i++) 
      {
        CoordXZ[0] = LsPar->pLaserPosX_MM[i];
        CoordXZ[1] = LsPar->pLaserPosZ_MM[i];
        
        FL_CPY_U8P_U8P(8, (U8 *)CoordXZ, pWrite[PN]); pWrite[PN] += 8;
      }
    }

    if(ParVal_DataMode == 5)
    {
      FL_CPY_U8P_U8P(4 * CountPoints, (U8 *)LsPar->pLaserPosX_MM, pWrite[PN]); pWrite[PN] += 4 * CountPoints;
      FL_CPY_U8P_U8P(4 * CountPoints, (U8 *)LsPar->pLaserPosZ_MM, pWrite[PN]); pWrite[PN] += 4 * CountPoints;
    }
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 8. State MODE_PRODUCT_RESULTS (ASCII)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  #ifdef PRODUCT_VERSION
  if( ParVal_DataMode == 8 )
  {
    I32   ProdResLen  = strlen(ProdPar->ProductResultString) + 1; 

    // set "Response Sub Header"
    SubHeaderLen = 4;
    SubHeader[0] = ProdResLen;
    FL_CPY_U8P_U8P(SubHeaderLen, (U8 *)SubHeader, pWrite[PN]); pWrite[PN] += SubHeaderLen;
    
    // response data
    FL_CPY_U8P_U8P(ProdResLen, (U8 *)ProdPar->ProductResultString, pWrite[PN]); pWrite[PN] += ProdResLen;
  }
  #endif


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // 9. State MODE_POINTLIST_MM + Product Results (x0 z0 / x1 z1 / x2 z2 / ...) + Product Results (ASCII)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  #ifdef PRODUCT_VERSION
  if( ParVal_DataMode == 9 )
  {
    I32   CountPoints = LsPar->LaserPoints;
    I32   ProdResLen  = strlen(ProdPar->ProductResultString) + 1; 
    float CoordXZ[2];
    
    // set "Response Sub Header"
    SubHeaderLen = 8;
    SubHeader[0] = CountPoints;
    SubHeader[1] = ProdResLen;
    FL_CPY_U8P_U8P(SubHeaderLen, (U8 *)SubHeader, pWrite[PN]); pWrite[PN] += SubHeaderLen;
    
    // write world coordinates into transfer buffer
    for(i=0; i<CountPoints; i++) 
    {
      CoordXZ[0] = LsPar->pLaserPosX_MM[i];
      CoordXZ[1] = LsPar->pLaserPosZ_MM[i];
      
      FL_CPY_U8P_U8P(8, (U8 *)CoordXZ, pWrite[PN]); pWrite[PN] += 8;
    }
    
    // transfer response data
    FL_CPY_U8P_U8P(ProdResLen, (U8 *)ProdPar->ProductResultString, pWrite[PN]); pWrite[PN] += ProdResLen;
    
    // print("SendCounter=%5d CountPoints=%4d ProdResLen=%2d Bytes=%4d\n", SendCounter[PN]-1, CountPoints, ProdResLen, 12+8*CountPoints+ProdResLen));
    // print("ProdRes: %s\n\n", (char *)(&pWrite[PN][-ProdResLen]));
    // print("ProdRes: %s Len=%3d  LastTwoChar: 0x%x 0x%x\n", ProdPar->ProductResultString, ProdResLen, (int)ProdPar->ProductResultString[ProdResLen-2], (int)ProdPar->ProductResultString[ProdResLen-1]);
  }
  #endif


  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // send MODE (VISIOM -> HOST)
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // print response string
  if (LsPar->PrintHostCommand & 0x04)
  {
    print("Cam->Host Port %d (SendData): ", PN);
    print("TcpSendCnt =%4d  CommandType=%d  DataMode=%2d  SendCounter=%4d  TimeStamp=%013lums  DataLen=%4d\n", TcpSendCnt[PN], 100, ParVal_DataMode, SendCounter[PN], LsPar->ImgTimeStampMs, (I32)(pWrite[PN] - pGlobalHeaderStart) - GLOBAL_HEAD_SIZE);
  }

  // create Gloabal Header
  SetGlobalHeader(pGlobalHeaderStart, (I32)(pWrite[PN] - pGlobalHeaderStart) - GLOBAL_HEAD_SIZE, 0, LsPar->PrintHostCommand & 0x08);

  
  PackCnt[PN]++;

  // transfer results (or build packages, if MODE = 4 or 5)
  // if ( ((ParVal_DataMode != 4) && (ParVal_DataMode != 5)) || (PackCnt[PN] >= LsPar->EthernetPackNr) || (ParVal_NbrLines == 0) )
  if ( (ParVal_NbrLines == 1) || ((PackCnt[PN] >= LsPar->EthernetPackNr) && (LsPar->EthernetPackNr)) || (ParVal_DataMode == 1) || (ParVal_DataMode == 6) )
  {
    I32 OptiSendSize;
    U40 ms, DiffMs;

    // reset package counter
    PackCnt[PN] = 0;

    // calc response len
    count = (I32)(pWrite[PN] - pSend[PN]);
    
    ms = GetTimeMs();

    while (count)
    {
      // 1460 Byte Send Mode is working not correct for the SmartShape Host
      // OptiSendSize = (count / 1460) * 1460;
      
      if(LsPar->DebugPrint == 9) print("SEND: Alias=0x%08x PN=%d TcpSendCnt=%5d TcpRecveCnt=%4d Mode=%d RestData=%7d Mode=%d NrLines=%2d\n", Alias, PN, TcpSendCnt[PN], TcpRecveCnt[PN], Mode, (int)(pWrite[PN] - pSend[PN]), ParVal_DataMode, ParVal_NbrLines);

      OptiSendSize = count;
      
      if(OptiSendSize > 44 * 1460) OptiSendSize = 44 * 1460;
      if(OptiSendSize == 0) OptiSendSize = count;

      // send full size in BlockingMode
      if(ParVal_DataMode == 1 || ParVal_DataMode == 6) OptiSendSize = count;

      SendSize = TCP_Send(Alias, (char *)pSend[PN], OptiSendSize);

      if(SendSize <  0) {if (LsPar->DebugPrint == 5) print("Error=%d TCP Send: Size=%d\n", SendSize, count); pSend[PN] = pWrite[PN] = Command; return -1;}
      if(SendSize == 0) {time_delay(1);}
      if(SendSize >  0) {ms = GetTimeMs();}

      count -= SendSize;
      pSend[PN] += SendSize;

      // stop send data, if more scans are requested
      if(LsPar->EthernetSendNoWait)
      {
        if( (ParVal_NbrLines != 1) && (LsPar->EthernetPackNr != 0) && (ParVal_DataMode != 1) && (ParVal_DataMode != 6)) break;
        // if( (ParVal_NbrLines != 0) && (LsPar->EthernetPackNr != 0) ) break;
      }

      // cancel while loop after n loops
      // if(LoopNr-- < 0) {pSend[PN] = pWrite[PN] = Command; count = 0;}
      
      // cancel while loop after n ms
      DiffMs = GetTimeMs() - ms;
      if (DiffMs > 500) break;

      if (LsPar->DebugPrint == 5) print("Main SendSize=%6d DiffMs=%3d Remain=%10d\n", SendSize, (I32)DiffMs, count);
    }

    if(count == 0) {pSend[PN] = pWrite[PN] = Command;}
  }


  return Error;
}

/************************************************************************************************************************/

I32 SetGlobalHeader (U8 *pWrite, I32 DataLength, I32 Error, I32 Debug)
{
  I32 i; 
  U32 SyncSeq      = GLOBAL_HEAD_SYNC;
  U32 Reserved     = 0;
  U32 Counter      = TcpSendCnt[PN];
  U32 TimeOutMs    = 0;
  U32 CheckSumData = 0, CheckSumHead = 0;
  
  
  // calc check sum HEADER of Counter, TimeOutMs, Reserved, Error and DataLength
  CheckSumHead = Counter ^ TimeOutMs ^ Reserved ^ Error ^ DataLength;
  CheckSumHead = ((CheckSumHead >> 24) & 0xFF) ^ ((CheckSumHead >> 16) & 0xFF) ^ ((CheckSumHead >> 8) & 0xFF) ^ (CheckSumHead & 0xFF);

  // calc check sum DATA
  CheckSumData = 0;
  for(i=0; i<DataLength; i++) CheckSumData = CheckSumData ^ pWrite[i + GLOBAL_HEAD_SIZE];

  // generate GLOBAL_HEAD_SIZE Bytes of Global Header
  FL_CPY_U8P_U8P(4, (U8 *)&SyncSeq,        pWrite); pWrite += 4;
  FL_CPY_U8P_U8P(4, (U8 *)&CheckSumHead,   pWrite); pWrite += 4;
  FL_CPY_U8P_U8P(4, (U8 *)&CheckSumData,   pWrite); pWrite += 4;
  FL_CPY_U8P_U8P(4, (U8 *)&Counter,        pWrite); pWrite += 4;
  FL_CPY_U8P_U8P(4, (U8 *)&TimeOutMs,      pWrite); pWrite += 4;
  FL_CPY_U8P_U8P(4, (U8 *)&Reserved,       pWrite); pWrite += 4;
  FL_CPY_U8P_U8P(4, (U8 *)&Error,          pWrite); pWrite += 4;
  FL_CPY_U8P_U8P(4, (U8 *)&DataLength,     pWrite); pWrite += 4;    

  if(Debug)  
  {
    Debug = 0;
    print("Cam->Host Port %d  (Header) : Sync=%X CSH=%02X CSD=%02X Cnt=%4d TOut=%3d Rsv=%d Err=%d DLen=%d\n", PN, SyncSeq, CheckSumHead, CheckSumData, Counter, TimeOutMs, Reserved, Error, DataLength);

    // Header
    if(Debug & 0x01) {for(i=0; i<GLOBAL_HEAD_SIZE;    i++) print("%02X ", pWrite[i-GLOBAL_HEAD_SIZE]); print("\n");}

    // Data ASCII
    if(Debug & 0x02) {for(i=0; i<min(DataLength, 80); i++) print("%c",    pWrite[i]); print("\n");}

    // Data Hex
    if(Debug & 0x04) {for(i=0; i<min(DataLength, 40); i++) print("%02X ", pWrite[i]); print("\n");}

    print("\n");
  }
  
  
  // increase sensor counter
  TcpSendCnt[PN]++;
 
  return 0;
}

/************************************************************************************************************************/

I32 GetGlobalHeader (U8 *pRead, I32 ReadLen, I32 *Error, I32 *TimeOutMs, I32 *CheckSumData, I32 *DataLength, I32 *Offset, I32 Debug)
{
  U32 i, Counter, CheckSumHead, CheckSumSend;
  U32 SyncSeq;
  U32 Reserved;
  U8  *pPos, *pEnd;
  
  // ini values
  SyncSeq       = 0;
  *Error        = 0;
  *TimeOutMs    = 0;
  *CheckSumData = 0;
  *DataLength   = 0;
  *Offset       = 0;
  
  pPos          = pRead;
  pEnd          = pRead + ReadLen - GLOBAL_HEAD_SIZE;
  

  // check ReadLen  
  if(ReadLen < GLOBAL_HEAD_SIZE) return ERR_PARM;
  
  
  // find SyncSeq in Data stream
  for(; pPos <= pEnd; pPos++)
  {
    FL_CPY_U8P_U8P(4, pPos, (U8 *)&SyncSeq); 

    if(SyncSeq == GLOBAL_HEAD_SYNC) break;
  }
  
    
  // did not find sync 
  if(SyncSeq != GLOBAL_HEAD_SYNC) 
  {
    *Offset = (I32)(pPos - pRead) - 1;
    return ERR_TCP_IP_SYNC;
  }
  
  
  // get global header info
  FL_CPY_U8P_U8P(4,  pPos, (U8 *)&SyncSeq     ); pPos += 4; 
  FL_CPY_U8P_U8P(4,  pPos, (U8 *)&CheckSumSend); pPos += 4;
  FL_CPY_U8P_U8P(4,  pPos, (U8 *) CheckSumData); pPos += 4;
  FL_CPY_U8P_U8P(4,  pPos, (U8 *)&Counter     ); pPos += 4;
  FL_CPY_U8P_U8P(4,  pPos, (U8 *) TimeOutMs   ); pPos += 4;
  FL_CPY_U8P_U8P(4,  pPos, (U8 *)&Reserved    ); pPos += 4;
  FL_CPY_U8P_U8P(4,  pPos, (U8 *) Error       ); pPos += 4;
  FL_CPY_U8P_U8P(4,  pPos, (U8 *) DataLength  ); pPos += 4;

  // varify check sum HEADER of Counter, TimeOutMs, Reserved, Error and DataLength
  CheckSumHead = Counter ^ *TimeOutMs ^ Reserved ^ *Error ^ *DataLength;
  CheckSumHead = ((CheckSumHead >> 24) & 0xFF) ^ ((CheckSumHead >> 16) & 0xFF) ^ ((CheckSumHead >> 8) & 0xFF) ^ (CheckSumHead & 0xFF);
  // if(CheckSumSend != CheckSumHead) Error = ERR_PARM;
  
  
  // data position after Global Header
  *Offset = (I32)(pPos - pRead);
  
  // set counter
  if(Counter ==               0) TcpRecveCnt[PN] = 0;
  
  
  if(Debug)  
  {
    Debug = 0;
    
    print("Host->Cam Port=%d  (Header) : Sync=%X CSH=%X CSD=%X Cnt=%3d [Intern=%3d] TOut=%3d Res=%d Err=%3d DLen=%4d Offs=%2d\n", PN, SyncSeq, CheckSumSend, *CheckSumData, Counter, TcpRecveCnt[PN], *TimeOutMs, Reserved, *Error, *DataLength, *Offset);

    // Header
    if(Debug & 0x01) {for(i=0; i<GLOBAL_HEAD_SIZE;     i++) print("%02X ", pPos[i-GLOBAL_HEAD_SIZE]); print("\n");}

    // Data ASCII
    if(Debug & 0x02) {for(i=0; i<min(*DataLength, 80); i++) print("%c",    pPos[i]); print("\n");}

    // Data Hex
    if(Debug & 0x04) {for(i=0; i<min(*DataLength, 40); i++) print("%02X ", pPos[i]); print("\n");}

    print("\n");
  }
    
  // check counter
  // if(Counter != TcpRecveCnt[PN]) return ERR_COUNTER; 
  if(Counter != TcpRecveCnt[PN]) print("Wrong Counter = %d instead of %d\n", Counter, TcpRecveCnt[PN]); // ???

  // increase host counter
  TcpRecveCnt[PN]++;
  
  return 0;
}

/************************************************************************************************************************/

I32 TCP_NewConnection (U32 Sock, U32 *Alias)
{
  /*
  return: 
    1: new connection ready
    0: no new connection made
   -1: VCRT Connection Socket Error
   -2: Connection Time Out Error
   -3: Accept Failed Error
  */

  I32  ConnectionTimeOut, result;
  U32  ContactingRemoteSocket, SockArray[2];


  // ConnectionTimeOut:  0 means indefinitely / -1 means do not block / else waiting time in ms
  ConnectionTimeOut = -1; 
  
  // ContactingRemoteSocket = VCRT_selectall(ConnectionTimeOut);

  SockArray[0] = Sock;
  SockArray[1] = *Alias;
  ContactingRemoteSocket = VCRT_selectset(SockArray, 2, ConnectionTimeOut);
  
  // problems with VCRT_selectall()
  if (ContactingRemoteSocket == VCRT_SOCKET_ERROR)
  {
    print ("VCRT Connection Socket Error = %d\n", ContactingRemoteSocket);
    return(-1);
  }

  // time out 
  if (ContactingRemoteSocket <= 0)
  {
    // print ("Connection Time Out Error\n");    
    return(-2); 
  }
 
  // Is there one who wants to connect Sock 
  if (Sock == ContactingRemoteSocket)
  {
    print("New Connection on Server Socket\n");

    // Close existing connection
    if (*Alias)
    {
      print("Close Old Socket\n");
      TCP_Close(*Alias);
      
      *Alias = 0;
    }

	  // get new connection socket 
    result = TCP_Accept(ContactingRemoteSocket, Alias);

    if (result)
    {
      print("Accept Failed Error 0x%x\n", VCRT_geterror(*Alias));
      *Alias = 0;
      return(-3);
    }
    else
    {
      print ("Accept new Connection at Sock=0x%x : Alias=0x%x\n", Sock, *Alias);
    }
    
    // new connection ready
    return (1);
  }
  
  // no new connection   
  return(0);
}

/************************************************************************************************************************/

I32 TCP_ReadData (U32 *Alias, U8 *pData, I32 MinLen, I32 MaxLen, I32 TimeOutMs, I32 *DataLen)
{
  I32  count=0, result=-1, DiffMs, ReadyString=0;
  U40  TimeMs;

  // get system time
  TimeMs = GetTimeMs();

  do
  {
    if (VCRT_geterror(*Alias) == VCRT_OK)
    {
      result = TCP_Receive(*Alias, (char *)&pData[count], MaxLen);

      if (VCRT_geterror(*Alias) != VCRT_OK)
      {
        if(*Alias) TCP_Close(*Alias);
        *Alias = 0;
        
        return(ERR_CMD_RECEIVE);
      }
    }
    else
    {
      if(*Alias) TCP_Close(*Alias);
      *Alias = 0;

      return(ERR_CMD_SOCKET);
    }   


    if(result >= 0)
    {
      count  += result;
      MaxLen -= result;

      if(count >= MinLen) ReadyString = 1;
    }
    else
    {
      if(count == 0) break;
    }

    DiffMs = GetTimeMs() - TimeMs;
  }
  while ( (DiffMs < TimeOutMs) && (ReadyString == 0) && (MaxLen > 0) );

  *DataLen = count;

  if (ReadyString) return(count);
  else             return(ERR_CMD_LESS_DATA);  
}

/************************************************************************************************************************/

#define TCP_IP_SOCKET

I32 TCP_Open (I32 Port, U32 *NewSock)
{
  unsigned  error, Sock;
  I32  Error;
//  unsigned  opt_value=1, opt_length=sizeof(unsigned);
  sockaddr_in  laddr;


  /* server services port */
  laddr.sin_family      = AF_INET;
  laddr.sin_port        = Port;        /* server listen at port */
  laddr.sin_addr.s_addr = INADDR_ANY;  /* every IP addr can connect */


  /* use TCP/IP protocoll and search for a free port */
  #ifdef TCP_IP_SOCKET
  Sock = socket_stream();
  #else
  Sock = socket_dgram();
  #endif

  if (Sock == VCRT_SOCKET_ERROR)
  {
    // print("Create stream socket failed\n");
    return(-1);
  }


  // max. send and receive buffer for fast ethernet transfer
  #if 1
  {
    #define SEND_BUFFER_SIZE    (0xFFFF) // max window size
    #define RECEIVE_BUFFER_SIZE (0xFFFF) // max window size

    I32 opt_value;

    // send buffer size
    opt_value = SEND_BUFFER_SIZE;
    Error = setsockopt(Sock, SOL_TCP, OPT_TBSIZE, &opt_value, sizeof(opt_value));
    if (Error != VCRT_OK)
    {
     print("Could not set socket option: OPT_TBSIZE!\n");
    }

    // receive buffer size
    opt_value = RECEIVE_BUFFER_SIZE;
    Error = setsockopt(Sock, SOL_TCP, OPT_RBSIZE, &opt_value, sizeof(opt_value));
    if (Error != VCRT_OK)
    {
     print("Could not set socket option: OPT_RBSIZE!\n");
    }
  }
  #endif

  #if 0
  {
    /* send data immediately */
    I32 opt_value = 1;

    Error = setsockopt(Sock, SOL_TCP, OPT_SEND_PUSH, &opt_value, sizeof(opt_value));
    if (Error != VCRT_OK)
    {
      print("ERROR: OPT_SEND_PUSH!\n");
    }
  }
  #endif

  #if 1
  {
    /* send data immediately (VCRT_MSG_NONBLOCK) */
    I32 opt_value = 1;

    Error = setsockopt(Sock, SOL_TCP, OPT_SEND_NOWAIT, &opt_value, sizeof(opt_value));
    if (Error != VCRT_OK)
    {
      print("ERROR: OPT_SEND_NOWAIT!\n");
    }
  }
  #endif




  #if 1
  {
    /* receive immediatly the data */
    I32 opt_value = 1;

    error = setsockopt(Sock, SOL_TCP, OPT_RECEIVE_NOWAIT, &opt_value, sizeof(opt_value)); 
    if (error != VCRT_OK)
    {
      print("Could not set socket option: OPT_RECEIVE_NOWAIT!\n");
    }
  }
  #endif


  /* connection between free port and selected SERVER_PORT. Now SERVER_PORT is free again for new connections. */
  error = bind(Sock, &laddr, sizeof(laddr));

  if (error != VCRT_OK)
  {
    // print("Stream bind failed - 0x%x\n", error);
    return(-2);
  }

  /* activate Sock for TCP/IP connections */
  #ifdef TCP_IP_SOCKET
  error = listen(Sock, 0);
  #endif

  if (error != VCRT_OK)
  {
    // print("Listen failed - 0x%x\n", error);
    return(-3);
  }

  // print("\nFast Ethernet port %d\n\n", laddr.sin_port);

  *NewSock = Sock;
  return(0);
}

/************************************************************************************************************************/

I32 TCP_Accept (U32 Sock, U32 *FromSock)
{
   uint_16  rlen;
   sockaddr_in  raddr;

   /* Connection requested; accept it */
   rlen = sizeof(raddr);

   *FromSock = accept(Sock, &raddr, &rlen);

   if (*FromSock == VCRT_SOCKET_ERROR)
   {
     return(-1);
   }

   return(0);
}

/************************************************************************************************************************/

I32 TCP_Send (U32 Sock, char *SendData, I32 SendBytes)
{
  I32 result;

#if 0
  {
   /* sends immediatly the data */
   I32 opt_value  = 1;
   I32 opt_length = sizeof(unsigned int);
   I32 error;
   error = setsockopt(Sock, SOL_TCP, OPT_SEND_PUSH, &opt_value, opt_length);
   if (error != VCRT_OK)
   {
   print("ERROR: OPT_SEND_PUSH!\n");
   }
 }
#endif

  if(getvar(ETHLINK))
  {
    result = send(Sock, (char *)SendData, SendBytes, 0);
  }
  else
  {
    result = VCRT_ERROR;
  }

  if(result==VCRT_ERROR)
  {
    return(-1);
  }

  return(result);
}

/************************************************************************************************************************/

I32 TCP_Receive (U32 Sock, char *ReceiveData, I32 ReceiveBytes)
{
   return( recv(Sock, (char *)ReceiveData, ReceiveBytes, 0) );
}

/************************************************************************************************************************/

I32 TCP_Close (U32 Sock)
{
  U32 Error;
  
  Error = shutdown(Sock, FLAG_CLOSE_TX);
  if(Error == 0) Error = shutdown(Sock, FLAG_ABORT_CONNECTION);
  
  return Error;
}

/************************************************************************************************************************/

U40 GetTimeMs (void)
{
  U40  TimeMs;
  static I40 SEC_Start = -1;
  
  // to avoid overflow at cameras with real time clock (SEC counts from year 1900)
  if(SEC_Start < 0) SEC_Start = getlvar(SEC);

  // disable interrupts
  disints();

  TimeMs = 1000L * (getlvar(SEC) - SEC_Start) + getvar(MSEC);

  // enable interrupts
  enaints();

  return TimeMs;
}

/************************************************************************************************************************/
/************************************************************************************************************************/


/************************************************************************************************************************/
/************************************************************************************************************************/
/************************************************************************************************************************/
